from BaseHTTPServer import BaseHTTPRequestHandler
from BaseHTTPServer import HTTPServer
import httplib
import requests
from urllib import quote
from urlparse import parse_qs
from urlparse import urlparse
import webbrowser

from cumulusci.oauth.exceptions import SalesforceOAuthError
from cumulusci.oauth.exceptions import RequestOauthTokenError

HTTP_HEADERS = {'Content-Type': 'application/x-www-form-urlencoded'}


class SalesforceOAuth2(object):

    def __init__(self, client_id, client_secret, callback_url, auth_site=None):
        if auth_site == None:
            auth_site = 'https://login.salesforce.com'
        self.client_id = client_id
        self.client_secret = client_secret
        self.callback_url = callback_url
        self.auth_site = auth_site

    def _request_token(self, request_data):
        url = self.auth_site + '/services/oauth2/token'
        data = {
            'client_id': self.client_id,
            'client_secret': self.client_secret,
        }
        data.update(request_data)
        response = requests.post(url, headers=HTTP_HEADERS, data=data)
        if response.status_code >= httplib.BAD_REQUEST:
            message = '{}: {}'.format(response.status_code, response.content)
            raise RequestOauthTokenError(message, response)
        return response

    def get_authorize_url(self, scope, prompt=None):
        url = self.auth_site + '/services/oauth2/authorize'
        url += '?response_type=code'
        url += '&client_id={}'.format(self.client_id)
        url += '&redirect_uri={}'.format(self.callback_url)
        url += '&scope={}'.format(quote(scope))
        if prompt:
            url += '&prompt={}'.format(quote(prompt))
        return url

    def get_token(self, code):
        data = {
            'grant_type': 'authorization_code',
            'redirect_uri': self.callback_url,
            'code': code,
        }
        return self._request_token(data)

    def refresh_token(self, refresh_token):
        data = {
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token,
        }
        return self._request_token(data)

    def revoke_token(self, current_token):
        url = self.auth_site + '/services/oauth2/revoke'
        data = {'token': quote(current_token)}
        response = requests.post(url, headers=HTTP_HEADERS, data=data)
        response.raise_for_status()
        return response


class OAuthCallbackHandler(BaseHTTPRequestHandler):
    parent = None

    def do_GET(self):
        args = parse_qs(urlparse(self.path).query, keep_blank_values=True)
        if 'error' in args:
            http_status = httplib.BAD_REQUEST
            http_body = 'error: {}\nerror description: {}'.format(
                args['error'], args['error_description'])
        else:
            http_status = httplib.OK
            http_body = 'OK'
            code = args['code']
            self.parent.response = self.parent.oauth_api.get_token(code)
            try:
                self.parent._check_response(self.parent.response)
            except SalesforceOAuthError as e:
                http_status = self.parent.response.status_code
                http_body = unicode(e)
        self.send_response(http_status)
        self.end_headers()
        self.wfile.write(http_body)


class CaptureSalesforceOAuth(object):

    def __init__(self, client_id, client_secret, callback_url, auth_site, scope):
        self.client_id = client_id
        self.client_secret = client_secret
        self.callback_url = callback_url
        self.auth_site = auth_site
        self.httpd = None
        self.oauth_api = self._get_oauth_api()
        self.response = None
        self.scope = scope
        self.httpd_timeout = 300

    def __call__(self):
        url = self._get_redirect_url()
        self._launch_browser(url)
        self._create_httpd()
        print (
            'Spawning HTTP server at {} '.format(self.callback_url) +
            'with timeout of {} seconds.\n'.format(self.httpd.timeout) +
            'If you are unable to log in to Salesforce you can ' +
            'press ctrl+c to kill the server and return to the command line.'
        )
        self.httpd.handle_request()
        self._check_response(self.response)
        return self.response.json()

    def _check_response(self, response):
        if response.status_code == httplib.OK:
            return
        raise SalesforceOAuthError('status_code:{} content:{}'.format(
            response.status_code, response.content))

    def _create_httpd(self):
        url_parts = urlparse(self.callback_url)
        server_address = (url_parts.hostname, url_parts.port)
        OAuthCallbackHandler.parent = self
        self.httpd = HTTPServer(server_address, OAuthCallbackHandler)
        self.httpd.timeout = self.httpd_timeout

    def _get_oauth_api(self):
        return SalesforceOAuth2(
            self.client_id,
            self.client_secret,
            self.callback_url,
            self.auth_site,
        )

    def _get_redirect_url(self):
        url = self.oauth_api.get_authorize_url(self.scope, prompt='login')
        response = requests.get(url)
        self._check_response(response)
        return url

    def _launch_browser(self, url):
        print 'Launching web browser for URL {}'.format(url)
        webbrowser.open(url, new=1)



import json
import os
import subprocess

from cumulusci.core.exceptions import CommandException
from cumulusci.core.tasks import BaseTask


class Command(BaseTask):

    task_options = {
        'command': {
            'description': 'The command to execute',
            'required': True,
        },
        'dir': {
            'description': 'If provided, the directory where the command should be run from.',
        },
        'env': {
            'description': 'Environment variables to set for command. Must be flat dict, either as python dict from YAML or as JSON string.',
        },
        'pass_env': {
            'description': 'If False, the current environment variables will not be passed to the child process.  Defaults to True',
            'required': True,
        },
    }

    def _init_options(self, kwargs):
        super(Command, self)._init_options(kwargs)
        if 'pass_env' not in self.options:
            self.options['pass_env'] = True
        if self.options['pass_env'] == 'False':
            self.options['pass_env'] = False
        if 'dir' not in self.options or not self.options['dir']:
            self.options['dir'] = '.'
        if 'env' not in self.options:
            self.options['env'] = {}
        else:
            try:
                self.options['env'] = json.loads(self.options['env'])
            except TypeError:
                # assume env is already dict
                pass

    def _run_task(self):
        env = self._get_env()
        self._run_command(env)
        
    def _get_env(self):
        if self.options['pass_env']:
            env = os.environ.copy()
        else:
            env = {}

        env.update(self.options['env'])
        return env

    def _process_output(self, line):
        self.logger.info(line.rstrip())
       
    def _handle_returncode(self, returncode): 
        if returncode:
            message = 'Return code: {}\nstderr: {}'.format(
                p.returncode,
                p.stderr,
            )
            self.logger.error(message)
            raise CommandException(message)

    def _run_command(self, env):
        p = subprocess.Popen(
            self.options['command'],
            stdout=subprocess.PIPE,
            bufsize=1,
            shell=True,
            executable='/bin/bash',
            env=env,
            cwd=self.options.get('dir'),
        )
        for line in iter(p.stdout.readline, ''):
            self._process_output(line)
        p.stdout.close()
        p.wait()
        if p.returncode:
            message = 'Return code: {}\nstderr: {}'.format(
                p.returncode,
                p.stderr,
            )
            self.logger.error(message)
            raise CommandException(message)

class SalesforceCommand(Command):
    """ A command that automatically gets a refreshed SF_ACCESS_TOKEN and SF_INSTANCE_URL passed as env vars """
    salesforce_task = True

    def _update_credentials(self):
        self.org_config.refresh_oauth_token(self.project_config.keychain.get_connected_app())

    def _get_env(self):
        env = super(SalesforceCommand, self)._get_env()
        env['SF_ACCESS_TOKEN'] = self.org_config.access_token
        env['SF_INSTANCE_URL'] = self.org_config.instance_url
        return env

task_options = Command.task_options.copy()
task_options['use_saucelabs'] = {
    'description': 'If True, use SauceLabs to run the tests.  The SauceLabs credentials will be fetched from the saucelabs service in the keychain and passed as environment variables to the command.  Defaults to False to run tests in the local browser.',
    'required': True,
}
class SalesforceBrowserTest(SalesforceCommand):
    """ A wrapper around browser test commands targetting a Salesforce org with support for running in local browser or on SauceLabs """

    task_options = task_options

    def _init_options(self, kwargs):
        super(SalesforceBrowserTest, self)._init_options(kwargs)
        if 'use_saucelabs' not in self.options or self.options['use_saucelabs'] == 'False':
            self.options['use_saucelabs'] = False
    
    def _get_env(self):
        env = super(SalesforceBrowserTest, self)._get_env()
        if self.options['use_saucelabs']:
            saucelabs = self.project_config.keychain.get_service('saucelabs')
            env['SAUCE_NAME'] = saucelabs.username
            env['SAUCE_KEY'] = saucelabs.api_key
            env['RUN_ON_SAUCE'] = 'True'
        else:
            env['RUN_LOCAL'] = 'True'
        return env



from cumulusci.tasks.github import BaseGithubTask
from cumulusci.tasks.release_notes.generator import GithubReleaseNotesGenerator
from cumulusci.tasks.release_notes.generator import PublishingGithubReleaseNotesGenerator

class GithubReleaseNotes(BaseGithubTask):

    task_options = {
        'tag': {
            'description': "The tag to generate release notes for.  Ex: release/1.2",
            'required': True,
        },
        'publish': {
            'description': "If True, publishes to the release matching the tag release notes were generated for.",
        },
        'last_tag': {
            'description': "Override the last release tag.  This is useful to generate release notes if you skipped one or more release",
        },
        'link_pr': {
            'description': 'If True, insert link to source pull request at end of each line.',
        }
    }
    
    def _run_task(self):
        github_info = {
            'github_owner': self.project_config.repo_owner,
            'github_repo': self.project_config.repo_name,
            'github_username': self.github_config.username,
            'github_password': self.github_config.password,
            'master_branch': self.project_config.project__git__default_branch,
            'prefix_beta': self.project_config.project__git__prefix_beta,
            'prefix_prod': self.project_config.project__git__prefix_release,
        }
        
        publish = self.options.get('publish', False) in (True, 'True', 'true')
        link_pr = self.options.get('link_pr', False) in (True, 'True', 'true')
        last_tag = self.options.get('last_tag', None)

        generator_class = GithubReleaseNotesGenerator

        if publish:
            generator_class = PublishingGithubReleaseNotesGenerator

        generator = generator_class(
            github_info,
            self.options['tag'],
            last_tag,
            link_pr,
        ) 
  
        release_notes = generator()
        self.logger.info('\n' + release_notes)



import json
import datetime
import functools
from simple_salesforce import Salesforce
from simple_salesforce import SalesforceMalformedRequest
from simple_salesforce.util import date_to_iso8601

from cumulusci.core.exceptions import CumulusCIException


def memoize(obj):
    cache = obj.cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cache:
            cache[key] = obj(*args, **kwargs)
        return cache[key]
    return memoizer

def batch_list(data, batch_size):
    batch_list = []
    batch_data = []
    for item in data:
        batch_data.append(item)
        if len(batch_data) == batch_size:
            batch_list.append(batch_data)
            batch_data = []
    if batch_data:
        batch_list.append(batch_data)
    return batch_list

class BasePushApiObject(object):
    def format_where(self, id_field, where=None):
        base_where = "%s = '%s'" % (id_field, self.sf_id)
        if where:
            where = '%s AND (%s)' % (base_where, where)
        else:
            where = base_where
        return where

# Push API Object Models

class MetadataPackage(BasePushApiObject):
    def __init__(self, push_api, name, sf_id=None, namespace=None):
        self.push_api = push_api
        self.sf_id = sf_id
        self.name = name
        self.namespace = namespace

    def get_package_versions(self, where=None, limit=None):
        where = self.format_where('MetadataPackageId', where)
        return self.push_api.get_package_versions(where, limit)

    def get_package_version_objs(self, where=None, limit=None):
        where = self.format_where('MetadataPackageId', where)
        return self.push_api.get_package_version_objs(where, limit)

    def get_package_versions_by_id(self, where=None, limit=None):
        where = self.format_where('MetadataPackageId', where)
        return self.push_api.get_package_versions_by_id(where, limit)

class MetadataPackageVersion(BasePushApiObject):
    def __init__(self, push_api, package, name, state, major, minor, patch, build, sf_id=None):
        self.push_api = push_api
        self.sf_id = sf_id
        self.package = package
        self.name = name
        self.state = state
        self.major = major
        self.minor = minor
        self.patch = patch
        self.build = build

    @property
    def version_number(self):
        parts = [str(self.major),str(self.minor)]
        if self.patch:
            parts.append(str(patch))
        version_number = '.'.join(parts)
        if self.state == 'Beta':
            version_number += ' (Beta %s)' % self.build
        return version_number

    def get_newer_released_version_objs(self, less_than_version=None):
        where = "MetadataPackageId = '%s' AND ReleaseState = 'Released' AND " % self.package.sf_id
        version_info = {'major': self.major, 'minor': self.minor, 'patch': self.patch}
        where += "(MajorVersion > %(major)s OR (MajorVersion = %(major)s AND MinorVersion > %(minor)s))" % version_info
        if self.patch:
            patch_where = " OR (MajorVersion = %(major)s AND MinorVersion = %(minor)s AND PatchVersion > %(patch))" % version_info
            where = where[:-1] + patch_where + where[-1:]

        if less_than_version:
            version_info = {
                'major': less_than_version.major,
                'minor': less_than_version.minor,
                'patch': less_than_version.patch
            }
            less_than_where = " AND (MajorVersion < %(major)s OR (MajorVersion = %(major)s AND MinorVersion < %(minor)s))" % version_info
            if less_than_version.patch:
                patch_where = " OR (MajorVersion = %(major)s AND MinorVersion = %(minor)s AND PatchVersion < %(patch))" % version_info
                less_than_where = less_than_where[:-1] + patch_where + less_than_where[-1:]
            where += less_than_where
    
        versions = self.package.get_package_version_objs(where)
        return versions

    def get_older_released_version_objs(self, greater_than_version=None):
        where = "MetadataPackageId = '%s' AND ReleaseState = 'Released' AND " % self.package.sf_id
        version_info = {'major': self.major, 'minor': self.minor, 'patch': self.patch}
        where += "(MajorVersion < %(major)s OR (MajorVersion = %(major)s AND MinorVersion < %(minor)s))" % version_info
        if self.patch:
            patch_where = " OR (MajorVersion = %(major)s AND MinorVersion = %(minor)s AND PatchVersion < %(patch))" % version_info
            where = where[:-1] + patch_where + where[-1:]
    
        if greater_than_version:
            version_info = {
                'major': greater_than_version.major,
                'minor': greater_than_version.minor,
                'patch': greater_than_version.patch
            }
            greater_than_where = " AND (MajorVersion > %(major)s OR (MajorVersion = %(major)s AND MinorVersion > %(minor)s))" % version_info
            if greater_than_version.patch:
                patch_where = " OR (MajorVersion = %(major)s AND MinorVersion = %(minor)s AND PatchVersion > %(patch))" % version_info
                greater_than_where = greater_than_where[:-1] + patch_where + greater_than_where[-1:]
            where += greater_than_where

        versions = self.package.get_package_version_objs(where)
        return versions

    def get_subscribers(self, where=None, limit=None):
        where = self.format_where('MetadataPackageVersionId', where)
        return self.push_api.get_subscribers(where, limit)

    def get_subscriber_objs(self, where=None, limit=None):
        where = self.format_where('MetadataPackageVersionId', where)
        return self.push_api.get_subscriber_objs(where, limit)

    def get_subscribers_by_org_key(self, where=None, limit=None):
        where = self.format_where('MetadataPackageVersionId', where)
        return self.push_api.get_subscribers_by_org_key(where, limit)

    def get_push_requests(self, where=None, limit=None):
        where = self.format_where('PackageVersionId', where)
        return self.push_api.get_push_requests(where, limit)

    def get_push_request_objs(self, where=None, limit=None):
        where = self.format_where('PackageVersionId', where)
        return self.push_api.get_push_request_objs(where, limit)

    def get_push_requests_by_id(self, where=None, limit=None):
        where = self.format_where('PackageVersionId', where)
        return self.push_api.get_push_requests_by_id(where, limit)

class PackagePushJob(BasePushApiObject):
    def __init__(self, push_api, request, org, status, sf_id=None):
        self.push_api = push_api
        self.request = request
        self.org = org
        self.status = status
        self.sf_id = sf_id

    def get_push_errors(self, where=None, limit=None):
        where = self.format_where('PackagePushJobId', where)
        return self.push_api.get_push_errors(where, limit)

    def get_push_error_objs(self, where=None, limit=None):
        where = self.format_where('PackagePushJobId', where)
        return self.push_api.get_push_error_objs(where, limit)

    def get_push_errors_by_id(self, where=None, limit=None):
        where = self.format_where('PackagePushJobId', where)
        return self.push_api.get_push_errors_by_id(where, limit)

class PackagePushError(BasePushApiObject):
    def __init__(self, push_api, job, severity, error_type, title, message, details, sf_id=None):
        self.push_api = push_api
        self.sf_id = sf_id
        self.job = job
        self.severity = severity
        self.error_type = error_type
        self.title = title
        self.message = message
        self.details = details

class PackagePushRequest(BasePushApiObject):
    def __init__(self, push_api, version, start_time, status, sf_id=None):
        self.push_api = push_api
        self.sf_id = sf_id
        self.version = version
        self.start_time = start_time
        self.status = status

    def get_push_jobs(self, where=None, limit=None):
        where = self.format_where('PackagePushRequestId', where)
        return self.push_api.get_push_jobs(where, limit)

    def get_push_job_objs(self, where=None, limit=None):
        where = self.format_where('PackagePushRequestId', where)
        return self.push_api.get_push_job_objs(where, limit)
        
    def get_push_jobs_by_id(self, where=None, limit=None):
        where = self.format_where('PackagePushRequestId', where)
        return self.push_api.get_push_jobs_by_id(where, limit)

class PackageSubscriber(object):
    def __init__(self, push_api, version, status, org_name, org_key, org_status, org_type, sf_id=None):
        self.push_api = push_api
        self.sf_id = sf_id
        self.version = version
        self.status = status
        self.org_name = org_name
        self.org_key = org_key
        self.org_status = org_status
        self.org_type = org_type

    def format_where(self, id_field, where=None):
        base_where = "%s = '%s'" % (id_field, self.org_key)
        if where:
            where = '%s AND (%s)' % (base_where, where)
        else:
            where = base_where
        return where

    def get_push_jobs(self, where=None, limit=None):
        where = self.format_where('SubscriberOrganizationKey', where)
        return self.push_api.get_push_jobs(where, limit)

    def get_push_job_objs(self, where=None, limit=None):
        where = self.format_where('SubscriberOrganizationKey', where)
        return self.push_api.get_push_job_objs(where, limit)

    def get_push_jobs_by_id(self, where=None, limit=None):
        where = self.format_where('SubscriberOrganizationKey', where)
        return self.push_api.get_push_jobs_by_id(where, limit)
       
 
class SalesforcePushApi(object):
    """ API Wrapper for the Salesforce Push API """

    def __init__(self, sf, logger, lazy=None, default_where=None, batch_size=None):
        self.sf = sf
        self.logger = logger

        if not lazy:
            lazy = []
        self.lazy = lazy

        if not default_where:
            default_where = {}
        self.default_where = default_where

        if not batch_size:
            batch_size = 200
        self.batch_size = batch_size

    def return_query_records(self, query):
        res = self.sf.query_all(query)
        if res['totalSize'] > 0:
            return res['records']
        else:
            return []

    def format_where_clause(self, where, obj=None):
        if obj and obj in self.default_where:
            default_where = self.default_where[obj]
            if where:
                where = '(%s) AND (%s)' % (default_where, where)
            else:
                where = '(%s)' % default_where
        if where:
            where = ' WHERE %s' % where
        else:
            where = ''
        return where

    def add_query_limit(self, query, limit):
        if not limit:
            return query

        return '%s LIMIT %s' % (query, limit)

    @memoize
    def get_packages(self, where=None, limit=None):
        where = self.format_where_clause(where)
        query = "SELECT id, name, namespaceprefix FROM MetadataPackage%s" % where
        query = self.add_query_limit(query, limit)
        return self.return_query_records(query)

    @memoize
    def get_package_objs(self, where=None, limit=None):
        package_objs = []
        for package in self.get_packages(where, limit):
            package_objs.append(
                MetadataPackage(
                    push_api = self,
                    sf_id=package['Id'], 
                    name=package['Name'], 
                    namespace=package['NamespacePrefix']
                )
            )
        return package_objs

    @memoize
    def get_packages_by_id(self, where=None, limit=None):
        packages = {}
        for package in self.get_package_objs(where, limit):
            packages[package.sf_id] = package
        return packages

    @memoize
    def get_package_versions(self, where=None, limit=None):
        where = self.format_where_clause(where)
        query = "SELECT Id, Name, MetadataPackageId, ReleaseState, MajorVersion, MinorVersion, PatchVersion, BuildNumber FROM MetadataPackageVersion%s ORDER BY MajorVersion DESC, MinorVersion DESC, PatchVersion, BuildNumber DESC" % where
        query = self.add_query_limit(query, limit)
        return self.return_query_records(query)

    @memoize
    def get_where_last_version(self, major=None, minor=None, beta=None):
        if beta:
            where = "ReleaseState = 'Beta'"
        else:
            where = "ReleaseState = 'Released'"
        if major:
            where += " AND MajorVersion=%s" % int(major)
        if minor:
            where += " AND MinorVersion=%s" % int(minor)
        return where
        
    @memoize
    def get_package_version_objs(self, where=None, limit=None):
        package_version_objs = []
        packages = self.get_packages_by_id()
        for package_version in self.get_package_versions(where, limit):
            package_version_objs.append(
                MetadataPackageVersion(
                    push_api = self,
                    name = package_version['Name'],
                    package = packages[package_version['MetadataPackageId']],
                    state = package_version['ReleaseState'],
                    major = package_version['MajorVersion'],
                    minor = package_version['MinorVersion'],
                    patch = package_version['PatchVersion'],
                    build = package_version['BuildNumber'],
                    sf_id = package_version['Id'],
                )
            )
        return package_version_objs

    @memoize
    def get_package_versions_by_id(self, where=None, limit=None):
        package_versions = {}
        for package_version in self.get_package_version_objs(where, limit):
            package_versions[package_version.sf_id] = package_version
        return package_versions

    @memoize
    def get_subscribers(self, where=None, limit=None):
        where = self.format_where_clause(where, obj='PackageSubscriber')
        query = "SELECT Id, MetadataPackageVersionId, InstalledStatus, OrgName, OrgKey, OrgStatus, OrgType from PackageSubscriber%s" % where
        query = self.add_query_limit(query, limit)
        return self.return_query_records(query)

    @memoize
    def get_subscriber_objs(self, where=None, limit=None):
        subscriber_objs = []
        package_versions = self.get_package_versions_by_id()
        for subscriber in self.get_subscribers(where, limit):
            subscriber_objs.append(
                PackageSubscriber(
                    push_api = self,
                    version = package_versions[subscriber['MetadataPackageVersionId']],
                    status = subscriber['InstalledStatus'],
                    org_name = subscriber['OrgName'],
                    org_key = subscriber['OrgKey'],
                    org_status = subscriber['OrgStatus'],
                    org_type = subscriber['OrgType'],
                    sf_id = subscriber['Id'],
                )
            )
        return subscriber_objs

    @memoize
    def get_subscribers_by_org_key(self, where=None, limit=None):
        subscribers = {}
        for subscriber in self.get_subscriber_objs(where, limit):
            subscribers[subscriber.org_key] = subscriber 
        return subscribers

    @memoize
    def get_push_requests(self, where=None, limit=None):
        where = self.format_where_clause(where, obj='PackagePushRequest')
        query = "SELECT Id, PackageVersionId, ScheduledStartTime, Status FROM PackagePushRequest%s ORDER BY ScheduledStartTime DESC" % where
        query = self.add_query_limit(query, limit)
        return self.return_query_records(query)

    @memoize
    def get_push_request_objs(self, where=None, limit=None):
        push_request_objs = []
        package_versions = self.get_package_versions_by_id()
        for push_request in self.get_push_requests(where, limit):
            push_request_objs.append(
                PackagePushRequest(
                    push_api = self,
                    version = package_versions[push_request['PackageVersionId']],
                    start_time=push_request['ScheduledStartTime'],
                    status = push_request['Status'],
                    sf_id = push_request['Id'],
                )
            )
        return push_request_objs

    @memoize
    def get_push_requests_by_id(self, where=None, limit=None):
        push_requests = {}
        for push_request in self.get_push_request_objs(where, limit):
            push_requests[push_request.sf_id] = push_request
        return push_requests

    @memoize
    def get_push_jobs(self, where=None, limit=None):
        where = self.format_where_clause(where)
        query = "SELECT Id, PackagePushRequestId, SubscriberOrganizationKey, Status FROM PackagePushJob%s" % where
        query = self.add_query_limit(query, limit)
        return self.return_query_records(query)

    @memoize
    def get_push_job_objs(self, where=None, limit=None):
        push_job_objs = []
        lazy = 'subscribers' in self.lazy
        if not lazy:
            subscriberorgs = self.get_subscribers_by_org_key()
        push_requests = self.get_push_requests_by_id()
        for push_job in self.get_push_jobs(where, limit):
            if lazy:
                orgs = self.get_subscriber_objs("OrgKey = '%s'" % push_job['SubscriberOrganizationKey'])
                if not orgs:
                    org = None
                else:
                    org = orgs[0]
            else:
                if push_job['SubscriberOrganizationKey'] not in subscriberorgs:
                    continue
                else:
                    org = subscriberorgs[push_job['SubscriberOrganizationKey']]
            push_job_objs.append(
                PackagePushJob(
                    push_api = self,
                    request = push_requests[push_job['PackagePushRequestId']],
                    org = org,
                    status = push_job['Status'],
                    sf_id = push_job['Id'],
                )
            )
        return push_job_objs

    @memoize
    def get_push_jobs_by_id(self, where=None, limit=None):
        push_jobs = {}
        for push_job in self.get_push_job_objs(where, limit):
            push_jobs[push_job.sf_id] = push_job
        return push_jobs

    @memoize
    def get_push_errors(self, where=None, limit=None):
        where = self.format_where_clause(where)
        query = "SELECT Id, PackagePushJobId, ErrorSeverity, ErrorType, ErrorTitle, ErrorMessage, ErrorDetails FROM PackagePushError%s" % where
        query = self.add_query_limit(query, limit)
        return self.return_query_records(query)

    @memoize
    def get_push_error_objs(self, where=None, limit=None):
        push_error_objs = []
        lazy = 'jobs' in self.lazy
        if not lazy:
            push_jobs = self.get_push_jobs_by_id()
        for push_error in self.get_push_errors(where, limit):
            if lazy:
                jobs = self.get_push_job_objs(where="Id = '%s'" % push_error['PackagePushJobId'])
                if jobs:
                    job = jobs[0]
                else:
                    job = None
             
            push_error_objs.append(
                PackagePushError(
                    push_api = self,
                    job = job,
                    severity = push_error['ErrorSeverity'],
                    error_type = push_error['ErrorType'],
                    title = push_error['ErrorTitle'],
                    message = push_error['ErrorMessage'],
                    details = push_error['ErrorDetails'],
                    sf_id = push_error['Id'],
                )
            )
        return push_error_objs

    @memoize
    def get_push_errors_by_id(self, where=None, limit=None):
        push_errors = {}
        for push_error in self.get_push_error_objs(where, limit):
            push_errors[push_error.sf_id] = push_error
        return push_errors

    def create_push_request(self, version, orgs, start=None):
        if not start:
            # Delay the push start by 15 minutes to allow manual review
            start = datetime.datetime.utcnow() + datetime.timedelta(minutes=15)

        # Create the request
        res = self.sf.PackagePushRequest.create({
            'PackageVersionId': version.sf_id,
            'ScheduledStartTime': start.isoformat(),
        })
        request_id = res['id']

        # remove duplicates
        n_orgs_pre = len(orgs)
        self.logger.info('Found {} orgs'.format(n_orgs_pre))
        orgs = set(orgs)
        if len(orgs) < n_orgs_pre:
            self.logger.warn('Removed {} duplicate orgs ({} remain)'.format(
                n_orgs_pre - len(orgs),
                len(orgs),
            ))

        # Schedule the orgs
        batches = batch_list(orgs, self.batch_size)
        scheduled_orgs = 0
        for batch_num, batch in enumerate(batches):
            self.logger.info(
                'Batch {} of {}: Attempting to add {} orgs'.format(
                    batch_num + 1,
                    len(batches),
                    len(batch),
                )
            )
            valid_batch = self._add_batch(batch, request_id)
            scheduled_orgs += len(valid_batch)
            self.logger.info(
                '{} orgs successfully added to batch'.format(len(valid_batch))
            )
        self.logger.info(
            'Push request {} is populated with {} orgs'.format(
                request_id,
                scheduled_orgs,
            )
        )
        return request_id

    def _add_batch(self, batch, request_id):

        # add orgs to batch data
        batch = set(batch)
        batch_data = {'records': []}
        for i, org in enumerate(batch):
            batch_data['records'].append({
                'attributes': {
                    'type': 'PackagePushJob',
                    'referenceId': 'org{}'.format(i),
                },
                'PackagePushRequestId': request_id,
                'SubscriberOrganizationKey': org,
            })

        # add batch to push request
        try:
            res = self.sf._call_salesforce(
                'POST',
                self.sf.base_url + 'composite/tree/PackagePushJob',
                data=json.dumps(batch_data),
            )
        except SalesforceMalformedRequest as e:
            invalid_orgs = set()
            for result in e.content['results']:
                for error in result['errors']:
                    if error['statusCode'] in [
                                'DUPLICATE_VALUE',
                                'INVALID_OPERATION',
                                'UNKNOWN_EXCEPTION',
                            ]:
                        org_id = self._get_org_id(
                            batch_data['records'],
                            result['referenceId'],
                        )
                        invalid_orgs.add(org_id)
                        self.logger.info('Skipping org {} - {}'.format(
                            org_id,
                            error['message'],
                        ))
                    else:
                        raise
            # remove invalid orgs and retry
            batch -= invalid_orgs
            if batch:
                self.logger.warn('Retrying batch without invalid orgs')
                batch = self._add_batch(batch, request_id)
            else:
                self.logger.error('Skipping batch (no valid orgs)')
        return batch

    def _get_org_id(self, records, ref_id):
        for record in records:
            if record['attributes']['referenceId'] == ref_id:
                return record['SubscriberOrganizationKey']

    def run_push_request(self, request_id):
        # Set the request to Pending status
        return self.sf.PackagePushRequest.update(request_id, {'Status': 'Pending'})



import json
import os
import sys
import webbrowser
import code

import click
from plaintable import Table
from rst2ansi import rst2ansi

import cumulusci
from cumulusci.core.config import ConnectedAppOAuthConfig
from cumulusci.core.config import FlowConfig
from cumulusci.core.config import OrgConfig
from cumulusci.core.config import ScratchOrgConfig
from cumulusci.core.config import ServiceConfig
from cumulusci.core.config import TaskConfig
from cumulusci.core.config import YamlGlobalConfig
from cumulusci.core.config import YamlProjectConfig
from cumulusci.core.exceptions import ApexTestException
from cumulusci.core.exceptions import FlowNotFoundError
from cumulusci.core.exceptions import KeychainConnectedAppNotFound
from cumulusci.core.exceptions import KeychainKeyNotFound
from cumulusci.salesforce_api.exceptions import MetadataApiError
from cumulusci.salesforce_api.exceptions import MetadataComponentFailure
from cumulusci.core.exceptions import NotInProject
from cumulusci.core.exceptions import ProjectConfigNotFound
from cumulusci.core.exceptions import ScratchOrgException
from cumulusci.core.exceptions import ServiceNotConfigured
from cumulusci.core.exceptions import TaskNotFoundError
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.exceptions import TaskRequiresSalesforceOrg
from cumulusci.core.utils import import_class
from cumulusci.utils import doc_task
from cumulusci.oauth.salesforce import CaptureSalesforceOAuth
from logger import init_logger


def pretty_dict(data):
    if not data:
        return ''
    return json.dumps(data, sort_keys=True, indent=4, separators=(',', ': '))

class CliConfig(object):
    def __init__(self):
        self.global_config = None
        self.project_config = None
        self.keychain = None

        init_logger()
        self._load_global_config()
        self._load_project_config()
        self._load_keychain()
        self._add_repo_to_path()

    def _add_repo_to_path(self):
        if self.project_config:
            sys.path.append(self.project_config.repo_root)

    def _load_global_config(self):
        try:
            self.global_config = YamlGlobalConfig()
        except NotInProject as e:
            raise click.UsageError(e.message)

    def _load_project_config(self):
        try:
            self.project_config = self.global_config.get_project_config()
        except ProjectConfigNotFound:
            pass
        except NotInProject as e:
            raise click.UsageError(e.message)

    def _load_keychain(self):
        self.keychain_key = os.environ.get('CUMULUSCI_KEY')
        if self.project_config:
            keychain_class = os.environ.get(
                'CUMULUSCI_KEYCHAIN_CLASS',
                self.project_config.cumulusci__keychain,
            )
            self.keychain_class = import_class(keychain_class)
            self.keychain = self.keychain_class(self.project_config, self.keychain_key)
            self.project_config.set_keychain(self.keychain)


try:
    CLI_CONFIG = CliConfig()
except click.UsageError as e:
    click.echo(e.message)
    sys.exit(1)


pass_config = click.make_pass_decorator(CliConfig, ensure=True)

def check_connected_app(config):
    check_keychain(config)
    if not config.keychain.get_connected_app():
        raise click.UsageError("Please use the 'org config_connected_app' command to configure the OAuth Connected App to use for this project's keychain")


def check_keychain(config):
    check_project_config(config)
    if config.project_config.keychain and config.project_config.keychain.encrypted and not config.keychain_key:
        raise click.UsageError('You must set the environment variable CUMULUSCI_KEY with the encryption key to be used for storing org credentials')

def check_project_config(config):
    if not config.project_config:
        raise click.UsageError('No project configuration found.  You can use the "project init" command to initilize the project for use with CumulusCI')

# Root command
@click.group('cli')
@pass_config
def cli(config):
    pass

@click.command(name='version', help='Print the current version of CumulusCI')
def version():
    click.echo(cumulusci.__version__)

@click.command(name='shell', help='Drop into a python shell')
@pass_config
@click.pass_context
def shell(ctx,config):
    code.interact(local=dict(globals(), **locals()))

# Top Level Groups
@click.group('project', help="Commands for interacting with project repository configurations")
@pass_config
def project(config):
    pass

@click.group('org', help="Commands for connecting and interacting with Salesforce orgs")
@pass_config
def org(config):
    pass

@click.group('task', help="Commands for finding and running tasks for a project")
@pass_config
def task(config):
    pass

@click.group('flow', help="Commands for finding and running flows for a project")
@pass_config
def flow(config):
    pass

@click.group('service',help="Commands for connecting services to the keychain")
@pass_config
def service(config):
    pass

cli.add_command(project)
cli.add_command(org)
cli.add_command(task)
cli.add_command(flow)
cli.add_command(version)
cli.add_command(shell)
cli.add_command(service)

# Commands for group: project

@click.command(name='init',
    help="Initialize a new project for use with the cumulusci toolbelt",
)
@click.option('--name',
    help="The project's package name",
    prompt=True,
)
@click.option('--package-name',
    help="The project's package name",
    prompt=True,
)
@click.option('--package-namespace',
    help="The project's package namespace",
    prompt=True,
)
@click.option('--package-api-version',
    help="The Salesforce API verson for the package",
    prompt=True,
    default=CLI_CONFIG.global_config.project__package__api_version,
)
@click.option('--git-prefix-feature',
    help="The branch prefix for all feature branches",
    prompt=True,
    default=CLI_CONFIG.global_config.project__git__prefix_feature,
)
@click.option('--git-default-branch',
    help="The default branch in the repository",
    prompt=True,
    default=CLI_CONFIG.global_config.project__git__default_branch,
)
@click.option('--git-prefix-beta',
    help="The tag prefix for beta release tags",
    prompt=True,
    default=CLI_CONFIG.global_config.project__git__prefix_beta,
)
@click.option('--git-prefix-release',
    help="The tag prefix for production release tags",
    prompt=True,
    default=CLI_CONFIG.global_config.project__git__prefix_release,
)
@click.option('--test-name-match',
    help="The SOQL format like query for selecting Apex tests.  % is wildcard",
    prompt=True,
    default=CLI_CONFIG.global_config.project__test__name_match,
)
@pass_config
def project_init(config, name, package_name, package_namespace, package_api_version, git_prefix_feature, git_default_branch, git_prefix_beta, git_prefix_release, test_name_match):
    if not os.path.isdir('.git'):
        click.echo("You are not in the root of a Git repository")

    if os.path.isfile('cumulusci.yml'):
        click.echo("This project already has a cumulusci.yml file")

    yml_config = []
    # project:
    yml_config.append('project:')
    yml_config.append('    name: {}'.format(name))

    #     package:
    package_config = []
    if package_name and package_name != config.global_config.project__package__name:
        package_config.append('        name: {}'.format(package_name))
    if package_namespace and package_namespace != config.global_config.project__package__namespace:
        package_config.append('        namespace: {}'.format(package_namespace))
    if package_api_version and package_api_version != config.global_config.project__package__api_version:
        package_config.append('        api_version: {}'.format(package_api_version))
    if package_config:
        yml_config.append('    package:')
        yml_config.extend(package_config)

    #     git:
    git_config = []
    if git_prefix_feature and git_prefix_feature != config.global_config.project__git__prefix_feature:
        git_config.append('        prefix_feature: {}'.format(git_prefix_feature))
    if git_default_branch and git_default_branch != config.global_config.project__git__default_branch:
        git_config.append('        default_branch: {}'.format(git_default_branch))
    if git_prefix_beta and git_prefix_beta != config.global_config.project__git__prefix_beta:
        git_config.append('        prefix_beta: {}'.format(git_prefix_beta))
    if git_prefix_release and git_prefix_release != config.global_config.project__git__prefix_release:
        git_config.append('        prefix_release: {}'.format(git_prefix_release))
    if git_config:
        yml_config.append('    git:')
        yml_config.extend(git_config)


    #     test:
    test_config = []
    if test_name_match and test_name_match != config.global_config.project__test__name_match:
        test_config.append('        name_match: {}'.format(test_name_match))
    if test_config:
        yml_config.append('    test:')
        yml_config.extend(test_config)

    yml_config.append('')
    
    with open('cumulusci.yml','w') as f_yml:
        f_yml.write('\n'.join(yml_config))

    click.echo("Your project is now initialized for use with CumulusCI")
    click.echo("You can use the project edit command to edit the project's config file")

@click.command(name='info', help="Display information about the current project's configuration")
@pass_config
def project_info(config):
    check_project_config(config)
    click.echo(pretty_dict(config.project_config.project))

@click.command(name='list', help="List projects and their locations")
@pass_config
def project_list(config):
    pass

@click.command(name='cd', help="Change to the project's directory")
@pass_config
def project_cd(config):
    pass

project.add_command(project_init)
project.add_command(project_info)
#project.add_command(project_list)
#project.add_command(project_cd)

# Commands for group: service
@click.command(name='list', help='List services available for configuration and use')
@pass_config
def service_list(config):
    headers = ['service','description','is_configured']
    data = []
    for serv,schema in config.project_config.services.iteritems():
        is_configured = ''
        if serv in config.keychain.list_services():
            is_configured = '* '
        data.append((serv,schema['description'],is_configured))
    table = Table(data, headers)
    click.echo(table)


class ConnectServiceCommand(click.MultiCommand):
    def list_commands(self, ctx):
        """ list the services that can be configured """
        config = ctx.ensure_object(CliConfig)
        return sorted(config.project_config.services.keys())

    def _build_param(self, attribute, details):
        req = details['required']
        return click.Option(('--{0}'.format(attribute),), prompt=req, required=req)

    def get_command(self, ctx, name):
        config = ctx.ensure_object(CliConfig)

        attributes = getattr(
            config.project_config, 
            'services__{0}__attributes'.format(name)
        ).iteritems()
        params = [self._build_param(attr,cnfg) for attr, cnfg in attributes]
        params.append(click.Option(('--project',),is_flag=True))

        @click.pass_context
        def callback(ctx,project=False,*args, **kwargs):
            check_keychain(config)
            serv_conf = dict((k, v) for k, v in kwargs.iteritems() if v!=None) # remove None values
            config.keychain.set_service(name, ServiceConfig(serv_conf), project)
            if project:
                click.echo('{0} is now configured for this project'.format(name))
            else:
                click.echo('{0} is now configured for global use'.format(name))

        ret = click.Command(name, params=params, callback=callback)
        return ret

@click.command(cls=ConnectServiceCommand,name='connect',help='Connect a CumulusCI task service')
@click.pass_context
def service_connect(ctx, *args, **kvargs):
    pass


@click.command(name='show',help='Show the details of a connected service')
@click.argument('service_name')
@pass_config
def service_show(config,service_name):
    check_keychain(config)
    try:
        service_config = config.keychain.get_service(service_name)
        click.echo(pretty_dict(service_config.config))
    except ServiceNotConfigured:
        click.echo('{0} is not configured for this project.  Use service connect {0} to configure.'.format(service_name))

service.add_command(service_connect)
service.add_command(service_list)
service.add_command(service_show)

# Commands for group: org
@click.command(name='browser', help="Opens a browser window and logs into the org using the stored OAuth credentials")
@click.argument('org_name')
@pass_config
def org_browser(config, org_name):
    check_connected_app(config)

    org_config = config.project_config.get_org(org_name)
    org_config.refresh_oauth_token(config.keychain.get_connected_app())

    webbrowser.open(org_config.start_url)

    # Save the org config in case it was modified
    config.keychain.set_org(org_name, org_config)

@click.command(name='connect', help="Connects a new org's credentials using OAuth Web Flow")
@click.argument('org_name')
@click.option('--sandbox', is_flag=True, help="If set, connects to a Salesforce sandbox org")
@click.option('--login-url', help='If set, login to this hostname.', default= 'https://login.salesforce.com')
@click.option('--global-org', help='Set True if org should be used by any project', is_flag=True)
@pass_config
def org_connect(config, org_name, sandbox, login_url, global_org):
    check_connected_app(config)

    connected_app = config.keychain.get_connected_app()
    if sandbox:
        login_url = 'https://test.salesforce.com'

    oauth_capture = CaptureSalesforceOAuth(
        client_id = connected_app.client_id,
        client_secret = connected_app.client_secret,
        callback_url = connected_app.callback_url,
        auth_site = login_url,
        scope = 'web full refresh_token'
    )
    oauth_dict = oauth_capture()
    org_config = OrgConfig(oauth_dict)
    org_config.load_userinfo()

    config.keychain.set_org(org_name, org_config, global_org)

@click.command(name='default', help="Sets an org as the default org for tasks and flows")
@click.argument('org_name')
@click.option('--unset', is_flag=True, help="Unset the org as the default org leaving no default org selected")
@pass_config
def org_default(config, org_name, unset):
    check_connected_app(config)

    if unset:
        org = config.keychain.unset_default_org()
        click.echo('{} is no longer the default org.  No default org set.'.format(org_name))
    else:
        org = config.keychain.set_default_org(org_name)
        click.echo('{} is now the default org'.format(org_name))


@click.command(name='info', help="Display information for a connected org")
@click.argument('org_name')
@pass_config
def org_info(config, org_name):
    check_connected_app(config)
    
    org_config = config.keychain.get_org(org_name)
    org_config.refresh_oauth_token(config.keychain.get_connected_app())

    click.echo(pretty_dict(org_config.config))

    # Save the org config in case it was modified
    config.keychain.set_org(org_name, org_config)

@click.command(name='list', help="Lists the connected orgs for the current project")
@pass_config
def org_list(config):
    check_connected_app(config)
    data = []
    headers = ['org','is_default']
    for org in config.project_config.list_orgs():
        org_config = config.project_config.get_org(org)
        if org_config.default:
            data.append((org, '*'))
        else:
            data.append((org, ''))
    table = Table(data, headers)
    click.echo(table)

@click.command(name='scratch', help="Connects a Salesforce DX Scratch Org to the keychain")
@click.argument('config_name')
@click.argument('org_name')
@click.option('--delete', is_flag=True, help="If set, triggers a deletion of the current scratch org.  This can be used to reset the org as the org configuration remains to regenerate the org on the next task run.")
@pass_config
def org_scratch(config, config_name, org_name, delete):
    check_connected_app(config)
  
    scratch_configs = getattr(config.project_config, 'orgs__scratch')
    if not scratch_configs:
        raise click.UsageError( 'No scratch org configs found in cumulusci.yml')
    scratch_config = scratch_configs.get(config_name)
    if not scratch_config:
        raise click.UsageError(
            'No scratch org config named {} found in the cumulusci.yml file'.format(config_name)
        )

    org_config = ScratchOrgConfig(scratch_config) 
    config.keychain.set_org(org_name, org_config)

@click.command(name='scratch_delete', help="Deletes a Salesforce DX Scratch Org leaving the config in the keychain for regeneration")
@click.argument('org_name')
@pass_config
def org_scratch_delete(config, org_name):
    check_connected_app(config)
    org_config = config.keychain.get_org(org_name)
    if not org_config.scratch:
        raise click.UsageError('Org {} is not a scratch org'.format(org_name))

    try:
        org_config.delete_org()
    except ScratchOrgException as e:
        exception = click.UsageError(e.message)

    config.keychain.set_org(org_name, org_config)

@click.command(name='connected_app', help="Displays the ConnectedApp info used for OAuth connections")
@pass_config
def org_connected_app(config):
    check_connected_app(config)
    click.echo(pretty_dict(config.keychain.get_connected_app().config))


@click.command(name='config_connected_app', help="Configures the connected app used for connecting to Salesforce orgs")
@click.option('--client_id', help="The Client ID from the connected app", prompt=True)
@click.option('--client_secret', help="The Client Secret from the connected app", prompt=True, hide_input=True)
@click.option('--callback_url', help="The callback_url configured on the Connected App", default='http://localhost:8080/callback')
@click.option('--project', help='Set if storing encrypted keychain file in project directory', is_flag=True)
@pass_config
def org_config_connected_app(config, client_id, client_secret, callback_url, project):
    check_keychain(config)
    app_config = ConnectedAppOAuthConfig()
    app_config.config = {
        'client_id': client_id,
        'client_secret': client_secret,
        'callback_url': callback_url,
    }
    config.keychain.set_connected_app(app_config, project)

org.add_command(org_browser)
org.add_command(org_config_connected_app)
org.add_command(org_connect)
org.add_command(org_connected_app)
org.add_command(org_default)
org.add_command(org_info)
org.add_command(org_list)
org.add_command(org_scratch)
org.add_command(org_scratch_delete)

# Commands for group: task
@click.command(name='list', help="List available tasks for the current context")
@pass_config
def task_list(config):
    check_project_config(config)
    data = []
    headers = ['task', 'description']
    for task in config.project_config.list_tasks():
        data.append((task['name'], task['description']))
    table = Table(data, headers)
    click.echo(table)

@click.command(name='doc', help="Exports RST format documentation for all tasks")
@pass_config
def task_doc(config):
    config_src = config.global_config
    
    for name, options in config_src.tasks.items():
        task_config = TaskConfig(options)
        doc = doc_task(name, task_config)
        click.echo(doc)
        click.echo('')

@click.command(name='info', help="Displays information for a task")
@click.argument('task_name')
@pass_config
def task_info(config, task_name):
    check_project_config(config)
    task_config = getattr(config.project_config, 'tasks__{}'.format(task_name))
    if not task_config:
        raise TaskNotFoundError('Task not found: {}'.format(task_name))

    task_config = TaskConfig(task_config)
    click.echo(rst2ansi(doc_task(task_name, task_config)))

@click.command(name='run', help="Runs a task")
@click.argument('task_name')
@click.option('--org', help="Specify the target org.  By default, runs against the current default org")
@click.option('-o', nargs=2, multiple=True, help="Pass task specific options for the task as '-o option value'.  You can specify more than one option by using -o more than once.")
@click.option('--debug', is_flag=True, help="Drops into pdb, the Python debugger, on an exception")
@pass_config
def task_run(config, task_name, org, o, debug):
    # Check environment
    check_keychain(config)

    # Get necessary configs
    if org:
        org_config = config.project_config.get_org(org)
    else:
        org, org_config = config.project_config.keychain.get_default_org()
    task_config = getattr(config.project_config, 'tasks__{}'.format(task_name))
    if not task_config:
        raise TaskNotFoundError('Task not found: {}'.format(task_name))

    # Get the class to look up options
    class_path = task_config.get('class_path')
    task_class = import_class(class_path)

    # Parse command line options and add to task config
    if o:
        if 'options' not in task_config:
            task_config['options'] = {}
        for option in o:
            name = option[0]
            value = option[1]

            # Validate the option
            if name not in task_class.task_options:
                raise click.UsageError(
                    'Option "{}" is not available for task {}'.format(
                        name,
                        task_name,
                    ),
                )

            # Override the option in the task config
            task_config['options'][name] = value

    task_config = TaskConfig(task_config)
    exception = None

    # Create and run the task
    try:
        task = task_class(config.project_config, task_config, org_config = org_config)
    except TaskRequiresSalesforceOrg as e:
        exception = click.UsageError('This task requires a salesforce org.  Use org default <name> to set a default org or pass the org name with the --org option')
    except TaskOptionsError as e:
        exception = click.UsageError(e.message)
    except Exception as e:
        if debug:
            import pdb
            import traceback
            traceback.print_exc()
            pdb.post_mortem()
        else:
            raise

    if not exception:
        try:
            task()
        except TaskOptionsError as e:
            exception = click.UsageError(e.message)
        except ApexTestException as e:
            exception = click.ClickException('Failed: ApexTestFailure')
        except MetadataComponentFailure as e:
            exception = click.ClickException('Failed: MetadataComponentFailure')
        except MetadataApiError as e:
            exception = click.ClickException('Failed: MetadataApiError')
        except Exception as e:
            if debug:
                import pdb
                import traceback
                traceback.print_exc()
                pdb.post_mortem()
            else:
                raise

    # Save the org config in case it was modified in the task
    if org and org_config:
        config.keychain.set_org(org, org_config)

    if exception:
        raise exception

# Add the task commands to the task group
task.add_command(task_doc)
task.add_command(task_info)
task.add_command(task_list)
task.add_command(task_run)

# Commands for group: flow
@click.command(name='list', help="List available flows for the current context")
@pass_config
def flow_list(config):
    check_project_config(config)
    data = []
    headers = ['flow', 'description']
    for flow in config.project_config.flows:
        description = getattr(config.project_config, 'flows__{}__description'.format(flow))
        data.append((flow, description))
    table = Table(data, headers)
    click.echo(table)

@click.command(name='info', help="Displays information for a flow")
@click.argument('flow_name')
@pass_config
def flow_info(config, flow_name):
    check_project_config(config)
    flow = getattr(config.project_config, 'flows__{}'.format(flow_name))
    if not flow:
        raise FlowNotFoundError('Flow not found: {}'.format(flow_name))
    click.echo(pretty_dict(flow))

@click.command(name='run', help="Runs a flow")
@click.argument('flow_name')
@click.option('--org', help="Specify the target org.  By default, runs against the current default org")
@click.option('--delete-org', is_flag=True, help="If set, deletes the scratch org after the flow completes")
@click.option('--debug', is_flag=True, help="Drops into pdb, the Python debugger, on an exception")
@pass_config
def flow_run(config, flow_name, org, delete_org, debug):
    # Check environment
    check_keychain(config)

    # Get necessary configs
    if org:
        org_config = config.project_config.get_org(org)
    else:
        org, org_config = config.project_config.keychain.get_default_org()

    if delete_org and not org_config.scratch:
        raise click.UsageError('--delete-org can only be used with a scratch org')

    flow = getattr(config.project_config, 'flows__{}'.format(flow_name))
    if not flow:
        raise FlowNotFoundError('Flow not found: {}'.format(flow_name))
    flow_config = FlowConfig(flow)
    if not flow_config.config:
        raise click.UsageError('No configuration found for flow {}'.format(flow_name))

    # Get the class to look up options
    class_path = flow_config.config.get('class_path', 'cumulusci.core.flows.BaseFlow')
    flow_class = import_class(class_path)

    exception = None

    # Create the flow and handle initialization exceptions
    try:
        flow = flow_class(config.project_config, flow_config, org_config)
    except TaskRequiresSalesforceOrg as e:
        exception = click.UsageError('This flow requires a salesforce org.  Use org default <name> to set a default org or pass the org name with the --org option')
    except TaskOptionsError as e:
        exception = click.UsageError(e.message)
    except Exception as e:
        if debug:
            import pdb
            import traceback
            traceback.print_exc()
            pdb.post_mortem()
        else:
            raise

    if not exception:
        # Run the flow and handle exceptions
        try:
            flow()
        except TaskOptionsError as e:
            exception = click.UsageError(e.message)
        except ApexTestException as e:
            exception = click.ClickException('Failed: ApexTestException')
        except MetadataComponentFailure as e:
            exception = click.ClickException('Failed: MetadataComponentFailure')
        except MetadataApiError as e:
            exception = click.ClickException('Failed: MetadataApiError')
        except Exception as e:
            if debug:
                import pdb
                import traceback
                traceback.print_exc()
                pdb.post_mortem()
            else:
                raise

    # Delete the scratch org if --delete-org was set
    if delete_org:
        try:
            org_config.delete_org()
        except Exception as e:
            click.echo('Scratch org deletion failed.  Ignoring the error below to complete the flow:')
            click.echo(e.message)

    # Save the org config in case it was modified in a task
    if org and org_config:
        config.keychain.set_org(org, org_config)

    if exception:
        raise exception

flow.add_command(flow_list)
flow.add_command(flow_info)
flow.add_command(flow_run)



from cumulusci.tasks.salesforce import BaseSalesforceBulkApiTask

import csv
import time
import hiyapyco
import xml.etree.ElementTree as ET

import datetime
import requests
import tempfile

from salesforce_bulk import CsvDictsAdapter

from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import Session
from sqlalchemy import create_engine
from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy import text
from sqlalchemy import types
from sqlalchemy import event
from StringIO import StringIO

# Create a custom sqlalchemy field type for sqlite datetime fields which are stored as integer of epoch time
class EpochType(types.TypeDecorator):
    impl = types.Integer

    epoch = datetime.datetime(1970, 1, 1, 0, 0, 0)

    def process_bind_param(self, value, dialect):
        return (value / 1000 - self.epoch).total_seconds()

    def process_result_value(self, value, dialect):
        return self.epoch + datetime.timedelta(seconds=value / 1000)

# Listen for sqlalchemy column_reflect event and map datetime fields to EpochType
@event.listens_for(Table, "column_reflect")
def setup_epoch(inspector, table, column_info):
    if isinstance(column_info['type'], types.DateTime):
        column_info['type'] = EpochType()

class DeleteData(BaseSalesforceBulkApiTask):

    task_options = {
        'objects': {
            'description': 'A list of objects to delete records from in order of deletion.  If passed via command line, use a comma separated string',
            'required': True,
        }
    }

    def _init_options(self, kwargs):
        super(DeleteData, self)._init_options(kwargs)
       
        # Split and trim objects string into a list if not already a list
        if not isinstance(self.options['objects'], list):
            self.options['objects'] = [obj.strip() for obj in self.options['objects'].split(',')]

    def _run_task(self):
        for obj in self.options['objects']:
            self.logger.info('Deleting all {} records'.format(obj))
            # Query for all record ids
            self.logger.info('  Querying for all {} objects'.format(obj))
            query_job = self.bulk.create_query_job(obj, contentType='CSV')
            batch = self.bulk.query(query_job, "select Id from {}".format(obj))
            while not self.bulk.is_batch_done(query_job, batch):
                time.sleep(10)
            self.bulk.close_job(query_job)

            delete_rows = []
            for row in self.bulk.get_batch_result_iter(query_job, batch, parse_csv=True):
                delete_rows.append(row)

            if not delete_rows:
                self.logger.info('  No {} objects found, skipping delete'.format(obj))
                continue

            # Delete the records
            delete_job = self.bulk.create_delete_job(obj, contentType='CSV')
            self.logger.info('  Deleting {} {} records'.format(len(delete_rows), obj))
            batch_num = 1
            for batch in self._upload_batch(delete_job, delete_rows):
                self.logger.info('    Uploaded batch {}'.format(batch))
                while not self.bulk.is_batch_done(delete_job, batch):
                    self.logger.info('      Checking status of batch'.format(batch_num))
                    time.sleep(10)
                self.logger.info('      Batch {} complete'.format(batch))
                batch_num += 1
            self.bulk.close_job(delete_job)

    def _split_batches(self, data, batch_size):
        """Yield successive n-sized chunks from l."""
        for i in range(0, len(data), batch_size):
            yield data[i:i + batch_size]

    def _upload_batch(self, job, data):
        # Split into batches
        batches = self._split_batches(data, 2500)

        uri = "{}/job/{}/batch".format(self.bulk.endpoint, job)
        headers = self.bulk.headers({"Content-Type": "text/csv"})
        for batch in batches:
            data = ['"Id"']
            data += ['"{}"'.format(record['Id']) for record in batch]
            data = '\n'.join(data)
            resp = requests.post(uri, data=data, headers=headers)
            content = resp.content

            if resp.status_code >= 400:
                self.bulk.raise_error(content, resp.status_code)

            tree = ET.fromstring(content)
            batch_id = tree.findtext("{%s}id" % self.bulk.jobNS)

            yield batch_id
        
class LoadData(BaseSalesforceBulkApiTask):

    task_options = {
        'database_url': {
            'description': 'The database url to a database containing the test data to load',
            'required': True,
        },
        'mapping': {
            'description': 'The path to a yaml file containing mappings of the database fields to Salesforce object fields',
            'required': True,
        },
    }

    def _run_task(self):
        self._init_mapping()
        self._init_db()

        for name, mapping in self.mapping.items():
            self.logger.info('Running Job: {}'.format(name))
            rows = self._get_batches(mapping)
            res = self._upload_batches(mapping, rows)

    def _create_job(self, mapping):
        action = mapping.get('action', 'insert')
        job_id = None

        if action == 'insert':
            job_id = self.bulk.create_insert_job(mapping['sf_object'], contentType='CSV')

        if not job_id:
            self.logger.error('  No handler for action type {}'.format(action))

        self.logger.info('  Created bulk job {}'.format(job_id))

        return job_id


    def _upload_batches(self, mapping, batches):
      
        job_id = None
        table = self.tables[mapping.get('table')]

        for batch, batch_rows in batches:
            if not job_id:
                # Create a job only once we have the first batch to load into it
                job_id = self._create_job(mapping)

            # Prepare the rows
            rows = CsvDictsAdapter(iter(batch))

            # Create the batch
            batch_id = self.bulk.post_bulk_batch(job_id, rows)
            self.logger.info('    Uploaded batch {}'.format(batch_id))
            while not self.bulk.is_batch_done(job_id, batch_id):
                self.logger.info('      Checking batch status...')
                time.sleep(10)
            
            # Wait for batch to complete
            res = self.bulk.wait_for_batch(job_id, batch_id)
            self.logger.info('      Batch {} complete'.format(batch_id))

            # salesforce_bulk is broken in fetching id results so do it manually
            results_url = '{}/job/{}/batch/{}/result'.format(self.bulk.endpoint, job_id, batch_id)
            headers = self.bulk.headers()
            resp = requests.get(results_url, headers=headers)
            csv_file = tempfile.TemporaryFile()
            csv_file.write(resp.content)
            csv_file.seek(0)
            reader = csv.DictReader(csv_file)

            # Write to the local Id column on the uploaded rows
            i = 0
            for result in reader:
                row = batch_rows[i]
                i += 1
                if result['Id']:
                    setattr(row, mapping['fields']['Id'], result['Id'])

            # Commit to the db
            self.session.commit()
                
        self.bulk.close_job(job_id)
 
        
    def _query_db(self, mapping):
        table = self.tables[mapping.get('table')]

        query = self.session.query(table)
        if 'filters' in mapping:
            filter_args = []
            for f in mapping['filters']:
                filter_args.append(text(f))
            query = query.filter(*filter_args)
        return query


    def _get_batches(self, mapping):
        action = mapping.get('action', 'insert')
        fields = mapping.get('fields', {}).copy()
        static = mapping.get('static', {})
        lookups = mapping.get('lookups', {})
        record_type = mapping.get('record_type')

        # Skip Id field on insert
        if action == 'insert' and 'Id' in fields:
            del fields['Id']

        # Build the list of fields to import
        import_fields = fields.keys() + static.keys() + lookups.keys()
        
        if record_type:
            import_fields.append('RecordTypeId')
            # query for record type by dev name
            record_type_id = None

        query = self._query_db(mapping)

        total_rows = 0
        batch_num = 1
        batch = []
        batch_rows = []
    
        for row in query:
            total_rows += 1

            # Get the row data from the mapping and database values
            csv_row = {}
            for key, value in fields.items():
                csv_row[key] = getattr(row, value)
            for key, value in static.items():
                csv_row[key] = value
            for key, lookup in lookups.items():
                lookup_table = lookup['table']
                kwargs = {lookup['join_field']: getattr(row, lookup['key_field'])}
                try:
                    res = self.session.query(self.tables[lookup['table']]).filter_by(**kwargs).one()
                    csv_row[key] = getattr(res, lookup['value_field'])
                except:
                    csv_row[key] = None
            if record_type:
                csv_row['RecordTypeId'] = record_type_id 

            # utf-8 encode row values
            for key, value in csv_row.items():
                if value:
                    if isinstance(value, datetime.datetime):
                        csv_row[key] = value.isoformat()
                    else:
                        try:
                            csv_row[key] = value.encode('utf8')
                        except AttributeError:
                            continue

            # Write to csv file
            batch.append(csv_row)
        
            batch_rows.append(row)

            # Slice into batches
            if len(batch) == 10000:
                self.logger.info('    Processing batch {}'.format(batch_num))
                yield batch, batch_rows

                # Start the next batch
                batch_num += 1
                batch = []
                batch_rows = []

        self.logger.info('  Prepared {} rows for import to {}'.format(total_rows, mapping['sf_object']))

        if batch:
            yield batch, batch_rows

    def _init_db(self):
        # initialize the DB engine
        self.engine = create_engine(self.options['database_url'])

        # initialize DB metadata
        self.metadata = MetaData()
        self.metadata.bind = self.engine

        # initialize the automap mapping
        self.base = automap_base(bind=self.engine, metadata=self.metadata)
        self.base.prepare(self.engine, reflect=True)

        # Loop through mappings and reflect each referenced table
        self.tables = {}
        for name, mapping in self.mapping.items():
            if 'table' in mapping and mapping['table'] not in self.tables:
                self.tables[mapping['table']] = self.base.classes[mapping['table']]

        # initialize the DB session
        self.session = Session(self.engine)

    def _init_mapping(self):
        self.mapping = hiyapyco.load(self.options['mapping'])



import httplib
import json

import requests

from cumulusci.core.exceptions import MrbelvedereError
from cumulusci.core.tasks import BaseTask


class BaseMrbelvedereTask(BaseTask):

    def _init_task(self):
        self.mrbelvedere_config = self.project_config.keychain.get_service(
            'mrbelvedere')


class MrbelvederePublish(BaseMrbelvedereTask):

    task_options = {
        'tag': {
            'description': 'The tag to publish to mrbelvedere',
            'required': True,
        },
    }

    def _init_task(self):
        super(MrbelvederePublish, self)._init_task()
        self.dependencies_url = '{}/{}/dependencies'.format(
            self.mrbelvedere_config.base_url,
            self.project_config.project__package__namespace,
        )
        if self.options['tag'].startswith(
                self.project_config.project__git__prefix_beta):
            self.dependencies_url += '/beta'
        self.current_dependencies = []

    def _run_task(self):
        self._get_current_dependencies()
        diffs = self._get_diffs()
        if diffs:
            response = requests.post(
                self.dependencies_url,
                data=json.dumps(diffs),
                headers={'Authorization': self.mrbelvedere_config.api_key},
            )
            if response.status_code >= httplib.BAD_REQUEST:
                raise MrbelvedereError('{}: {}'.format(
                    response.status_code, response.content))
            self.logger.info(response.content)

    def _get_current_dependencies(self):
        response = requests.get(self.dependencies_url)
        self.current_dependencies = json.loads(response.content)

    def _get_diffs(self):
        # flatten dependency tree, resolve duplicates
        dependencies = self._clean_dependencies(
            self.project_config.project__dependencies)
        # determine diffs vs current dependencies
        diffs = []
        for current_dependency in self.current_dependencies:
            match = False
            for dependency in dependencies:
                if dependency['namespace'] == current_dependency['namespace']:
                    match = True
                    if dependency['number'] == current_dependency['number']:
                        self.logger.info('No change for %s',
                                         current_dependency['namespace'])
                    else:
                        self.logger.info('Changing %s from version %s to %g',
                            dependency['namespace'],
                            current_dependency['number'],
                            dependency['number'],
                        )
                        diffs.append(dependency)
                    break
            if not match:
                self.logger.info('No change for %s',
                                 current_dependency['namespace'])
        diffs.append({
            'namespace': self.project_config.project__package__namespace,
            'number': self.project_config.get_version_for_tag(
                self.options['tag']),
        })
        return diffs

    def _clean_dependencies(self, dependency_list):
        cleaned_dependencies = []
        if not dependency_list:
            return cleaned_dependencies
        for dependency in dependency_list:
            if 'dependencies' in dependency:
                cleaned_dependencies.extend(
                    self._clean_dependencies(dependency['dependencies']))
            # update version
            exists = False
            for item in cleaned_dependencies:
                if item['namespace'] == dependency['namespace']:
                    exists = True
                    if item['number'] < dependency['version']:
                        item['number'] = str(dependency['version'])
                    break
            # add new entry
            if not exists:
                cleaned_dependencies.append({
                    'namespace': dependency['namespace'],
                    'number': str(dependency['version']),
                })
        return cleaned_dependencies



from base64 import b64encode
from zipfile import ZipFile
from tempfile import TemporaryFile
from xml.sax.saxutils import escape

INSTALLED_PACKAGE_PACKAGE_XML = """<?xml version="1.0" encoding="utf-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
  <types>
    <members>{}</members>
    <name>InstalledPackage</name>
  </types>
<version>33.0</version>
</Package>"""

EMPTY_PACKAGE_XML = """<?xml version="1.0" encoding="utf-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
<version>33.0</version>
</Package>"""

FULL_NAME_PACKAGE_XML = """<?xml version="1.0" encoding="utf-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
<fullName>{}</fullName>
<version>{}</version>
</Package>"""

INSTALLED_PACKAGE = """<?xml version="1.0" encoding="UTF-8"?>
<InstalledPackage xmlns="http://soap.sforce.com/2006/04/metadata">
  <versionNumber>{}</versionNumber>
</InstalledPackage>"""

class BasePackageZipBuilder(object):

    def __call__(self):
        self._open_zip()
        self._populate_zip()
        return self._encode_zip()

    def _open_zip(self):
        self.zip_file = TemporaryFile()
        self.zip= ZipFile(self.zip_file, 'w')

    def _populate_zip(self):
        raise NotImplementedError('Subclasses need to provide their own implementation')

    def _write_package_xml(self, package_xml):
        self.zip.writestr('package.xml', package_xml)

    def _write_file(self, path, content):
        self.zip.writestr(path, content)

    def _encode_zip(self):
        self.zip.close()
        self.zip_file.seek(0)
        return b64encode(self.zip_file.read())

class CreatePackageZipBuilder(BasePackageZipBuilder):

    def __init__(self, name, api_version):
        if not name:
            raise ValueError('You must provide a name to create a package')
        if not api_version:
            raise ValueError('You must provide an api_version to create a package')
        self.name= name
        self.api_version= api_version

    def _populate_zip(self):
        package_xml = FULL_NAME_PACKAGE_XML.format(escape(self.name), self.api_version)
        self._write_package_xml(package_xml)

class InstallPackageZipBuilder(BasePackageZipBuilder):

    def __init__(self, namespace, version):
        if not namespace:
            raise ValueError('You must provide a namespace to install a package')
        if not version:
            raise ValueError('You must provide a version to install a package')
        self.namespace = namespace
        self.version = version

    def _populate_zip(self):
        package_xml = INSTALLED_PACKAGE_PACKAGE_XML.format(self.namespace)
        self._write_package_xml(package_xml)

        installed_package = INSTALLED_PACKAGE.format(self.version)
        self._write_file(
            'installedPackages/{}.installedPackage'.format(self.namespace),
            installed_package
        )

class DestructiveChangesZipBuilder(BasePackageZipBuilder):

    def __init__(self, destructive_changes):
        self.destructive_changes = destructive_changes

    def _populate_zip(self): 
        self._write_package_xml(EMPTY_PACKAGE_XML)
        self._write_file('destructiveChanges.xml', self.destructive_changes)

class UninstallPackageZipBuilder(DestructiveChangesZipBuilder):

    def __init__(self, namespace):
        if not namespace:
            raise ValueError('You must provide a namespace to install a package')
        self.namespace = namespace
        self.destructive_changes = INSTALLED_PACKAGE_PACKAGE_XML.format(self.namespace)



import unittest

from mock import MagicMock
from mock import patch
import responses

from cumulusci.core.config import BaseGlobalConfig
from cumulusci.core.config import BaseProjectConfig
from cumulusci.core.config import ConnectedAppOAuthConfig
from cumulusci.core.config import OrgConfig
from cumulusci.core.config import TaskConfig
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.tasks.salesforce import BaseSalesforceToolingApiTask
from cumulusci.tasks.salesforce import RunApexTests
from cumulusci.tasks.salesforce import RunApexTestsDebug


@patch('cumulusci.tasks.salesforce.BaseSalesforceTask._update_credentials',
    MagicMock(return_value=None))
class TestBaseSalesforceToolingApiTask(unittest.TestCase):

    def setUp(self):
        self.api_version = 36.0
        self.global_config = BaseGlobalConfig(
            {'project': {'package': {'api_version': self.api_version}}})
        self.project_config = BaseProjectConfig(self.global_config)
        self.project_config.config['project'] = {
            'package': {
                'api_version': self.api_version,
            }
        }
        self.task_config = TaskConfig()
        self.org_config = OrgConfig({
            'instance_url': 'example.com',
            'access_token': 'abc123',
        })
        self.base_tooling_url = 'https://{}/services/data/v{}/tooling/'.format(
            self.org_config.instance_url, self.api_version)

    def test_get_tooling_object(self):
        task = BaseSalesforceToolingApiTask(
            self.project_config, self.task_config, self.org_config)
        obj = task._get_tooling_object('TestObject')
        url = self.base_tooling_url + 'sobjects/TestObject/'
        self.assertEqual(obj.base_url, url)


@patch('cumulusci.tasks.salesforce.BaseSalesforceTask._update_credentials',
    MagicMock(return_value=None))
class TestRunApexTests(unittest.TestCase):

    def setUp(self):
        self.api_version = 38.0
        self.global_config = BaseGlobalConfig(
            {'project': {'api_version': self.api_version}})
        self.task_config = TaskConfig()
        self.task_config.config['options'] = {
            'junit_output': 'results_junit.xml',
            'poll_interval': 1,
            'test_name_match': '%_TEST',
        }
        self.project_config = BaseProjectConfig(self.global_config)
        self.project_config.config['project'] = {'package': {
            'api_version': self.api_version}}
        keychain = BaseProjectKeychain(self.project_config, '')
        app_config = ConnectedAppOAuthConfig()
        keychain.set_connected_app(app_config)
        self.project_config.set_keychain(keychain)
        self.org_config = OrgConfig({
            'id': 'foo/1',
            'instance_url': 'example.com',
            'access_token': 'abc123',
        })
        self.base_tooling_url = 'https://{}/services/data/v{}/tooling/'.format(
            self.org_config.instance_url, self.api_version)

    def _mock_apex_class_query(self):
        url = (self.base_tooling_url + 'query/?q=SELECT+Id%2C+Name+' +
            'FROM+ApexClass+WHERE+NamespacePrefix+%3D+null' +
            '+AND+%28Name+LIKE+%27%25_TEST%27%29')
        expected_response = {
            'done': True,
            'records': [{'Id': 1, 'Name': 'TestClass_TEST'}],
            'totalSize': 1,
        }
        responses.add(responses.GET, url, match_querystring=True,
            json=expected_response)

    def _mock_get_test_results(self):
        url = (self.base_tooling_url + 'query/?q=SELECT+StackTrace%2C+' +
            'Message%2C+ApexLogId%2C+AsyncApexJobId%2C+MethodName%2C+' +
            'Outcome%2C+ApexClassId%2C+TestTimestamp+FROM+ApexTestResult+' +
            'WHERE+AsyncApexJobId+%3D+%27JOB_ID1234567%27')
        expected_response = {
            'done': True,
            'records': [{
                'ApexClassId': 1,
                'ApexLogId': 1,
                'Id': 1,
                'Message': 'Test passed',
                'MethodName': 'TestMethod',
                'Name': 'TestClass_TEST',
                'Outcome': 'Pass',
                'StackTrace': '1. ParentFunction\n2. ChildFunction',
                'Status': 'Completed',
            }],
        }
        responses.add(responses.GET, url, match_querystring=True,
            json=expected_response)

    def _mock_tests_complete(self):
        url = (self.base_tooling_url + 'query/?q=SELECT+Id%2C+Status%2C+' +
            'ApexClassId+FROM+ApexTestQueueItem+WHERE+ParentJobId+%3D+%27' +
            'JOB_ID1234567%27')
        expected_response = {
            'done': True,
            'records': [{'Status': 'Completed'}],
        }
        responses.add(responses.GET, url, match_querystring=True,
            json=expected_response)

    def _mock_run_tests(self):
        url = self.base_tooling_url + 'runTestsAsynchronous'
        expected_response = 'JOB_ID1234567'
        responses.add(responses.POST, url, json=expected_response)

    @responses.activate
    def test_run_task(self):
        self._mock_apex_class_query()
        self._mock_run_tests()
        self._mock_tests_complete()
        self._mock_get_test_results()
        task = RunApexTests(
            self.project_config, self.task_config, self.org_config)
        task()
        self.assertEqual(len(responses.calls), 4)


@patch('cumulusci.tasks.salesforce.BaseSalesforceTask._update_credentials',
    MagicMock(return_value=None))
class TestRunApexTestsDebug(TestRunApexTests):

    def setUp(self):
        super(TestRunApexTestsDebug, self).setUp()
        self.task_config.config['json_output'] = 'results.json'

    def _mock_create_debug_level(self):
        url = self.base_tooling_url + 'sobjects/DebugLevel/'
        expected_response = {
            'id': 1,
        }
        responses.add(responses.POST, url, json=expected_response)

    def _mock_create_trace_flag(self):
        url = self.base_tooling_url + 'sobjects/TraceFlag/'
        expected_response = {
            'id': 1,
        }
        responses.add(responses.POST, url, json=expected_response)

    def _mock_delete_debug_levels(self):
        url = self.base_tooling_url + 'sobjects/DebugLevel/1'
        responses.add(responses.DELETE, url)

    def _mock_delete_trace_flags(self):
        url = self.base_tooling_url + 'sobjects/TraceFlag/1'
        responses.add(responses.DELETE, url)

    def _mock_get_duration(self):
        url = (self.base_tooling_url + 'query/?q=SELECT+Id%2C+' +
            'Application%2C+DurationMilliseconds%2C+Location%2C+LogLength%2C' +
            '+LogUserId%2C+Operation%2C+Request%2C+StartTime%2C+Status+' +
            'from+ApexLog+where+Id+in+%28%271%27%29')
        expected_response = {
            'done': True,
            'records': [{'Id': 1, 'DurationMilliseconds': 1}],
            'totalSize': 1,
        }
        responses.add(responses.GET, url, match_querystring=True,
            json=expected_response)

    def _mock_get_log_body(self):
        url = self.base_tooling_url + 'sobjects/ApexLog/1/Body'
        expected_response = {
            'foo': 'bar',
        }
        responses.add(responses.GET, url, json=expected_response)

    def _mock_get_trace_flags(self):
        url = self.base_tooling_url + 'query/?q=Select+Id+from+TraceFlag+Where+TracedEntityId+%3D+%271%27'
        expected_response = {
            'records': [{'Id': 1}],
            'totalSize': 1,
        }
        responses.add(responses.GET, url, match_querystring=True,
            json=expected_response)

    def _mock_get_debug_levels(self):
        url = self.base_tooling_url + 'query/?q=Select+Id+from+DebugLevel'
        expected_response = {
            'records': [{'Id': 1}],
            'totalSize': 1,
        }
        responses.add(responses.GET, url, match_querystring=True,
            json=expected_response)

    @responses.activate
    def test_run_task(self):
        self._mock_apex_class_query()
        self._mock_get_trace_flags()
        self._mock_get_debug_levels()
        self._mock_delete_debug_levels()
        self._mock_delete_trace_flags()
        self._mock_create_debug_level()
        self._mock_create_trace_flag()
        self._mock_run_tests()
        self._mock_tests_complete()
        self._mock_get_test_results()
        self._mock_get_duration()
        self._mock_get_log_body()
        task = RunApexTestsDebug(
            self.project_config, self.task_config, self.org_config)
        task()
        self.assertEqual(len(responses.calls), 13)



import time

from datetime import datetime
from github3 import login
from github3 import GitHubError

from cumulusci.core.exceptions import GithubException
from cumulusci.core.tasks import BaseTask

class BaseGithubTask(BaseTask):

    def _init_task(self):
        self.github_config = self.project_config.keychain.get_service('github')
        self.github = login(
            username=self.github_config.username,
            password=self.github_config.password,
        )

    def get_repo(self):
        return self.github.repository(
            self.project_config.repo_owner,
            self.project_config.repo_name,
        )

class PullRequests(BaseGithubTask):
    
    def _run_task(self):
        repo = self.get_repo()
        for pr in repo.iter_pulls(state='open'):
            self.logger.info('#{}: {}'.format(pr.number, pr.title))

class CloneTag(BaseGithubTask):

    task_options = {
        'src_tag': {
            'description': 'The source tag to clone.  Ex: beta/1.0-Beta_2',
            'required': True,
        },
        'tag': {
            'description': 'The new tag to create by cloning the src tag.  Ex: release/1.0',
            'required': True,
        },
    }
    
    def _run_task(self):
        repo = self.get_repo()
        ref = repo.ref('tags/{}'.format(self.options['src_tag']))
        src_tag = repo.tag(ref.object.sha)
        if not src_tag:
            message = 'Tag {} not found'.format(self.options['src_tag'])
            logger.error(message)
            raise GithubException(message)

        tag = repo.create_tag(
            tag = self.options['tag'],
            message = 'Cloned from {}'.format(self.options['src_tag']),
            sha = src_tag.sha,
            obj_type = 'commit',
            tagger = {
                'name': self.github_config.username,
                'email': self.github_config.email,
                'date': '{}Z'.format(datetime.now().isoformat()),
            },
        )
        self.logger.info('Tag {} created by cloning {}'.format(self.options['tag'], self.options['src_tag']))

        return tag

class CreateRelease(BaseGithubTask):

    task_options = {
        'version': {
            'description': "The managed package version number.  Ex: 1.2",
            'required': True,
        },
        'message': {
            'description': "The message to attach to the created git tag",
        },
        'commit': {
            'description': "Override the commit used to create the release.  Defaults to the current local HEAD commit",
        },
    }
    
    def _run_task(self):
        repo = self.get_repo()

        for release in repo.iter_releases():
            if release.name == self.options['version']:
                message = 'Release {} already exists at {}'.format(release.name, release.html_url)
                self.logger.error(message)
                return GithubException(message)

        commit = self.options.get('commit', self.project_config.repo_commit)
        if not commit:
            message = 'Could not detect the current commit from the local repo'
            self.logger.error(message)
            return GithubException(message)

        version = self.options['version']
        self.tag_name = self.project_config.get_tag_for_version(version)

        ref = repo.ref('tags/{}'.format(self.tag_name))

        if not ref:
            # Create the annotated tag
            tag = repo.create_tag(
                tag = self.tag_name,
                message = 'Release of version {}'.format(version),
                sha = commit,
                obj_type = 'commit',
                tagger = {
                    'name': self.github_config.username,
                    'email': self.github_config.email,
                    'date': '{}Z'.format(datetime.now().isoformat()),
                },
                lightweight = False,
            )
    
            # Get the ref created from the previous call that for some reason creates
            # a ref to the commit sha rather than the tag sha.  Delete the ref so we
            # can create the right one.  FIXME: Is this a bug in github3.py?
            ref = repo.ref('tags/{}'.format(self.tag_name))
            if ref:
                ref.delete()
    
            # Create the ref linking to the tag
            ref = repo.create_ref(
                ref = 'refs/tags/{}'.format(self.tag_name),
                sha = tag.sha,
            )

            # Sleep for Github to catch up with the fact that the tag actually exists!
            time.sleep(3)

        prerelease = 'Beta' in version

        # Create the Github Release
        release = repo.create_release(
            tag_name = self.tag_name,
            name = version,
            prerelease = prerelease,
        )
        self.return_values = {'tag_name': self.tag_name}
        self.logger.info('Created release {} at {}'.format(release.name, release.html_url))

class MergeBranch(BaseGithubTask):

    task_options = {
        'commit': {
            'description': "The commit to merge into feature branches.  Defaults to the current head commit.",
        },
        'source_branch': {
            'description': "The source branch to merge from.  Defaults to project__git__default_branch.",
        },
        'branch_prefix': {
            'description': "The prefix of branches that should receive the merge.  Defaults to project__git__prefix_feature",
        },
    }

    def _run_task(self):
        repo = self.get_repo()

        commit = self.options.get('commit', self.project_config.repo_commit)
        branch_prefix = self.options.get('branch_prefix', self.project_config.project__git__prefix_feature)
        source_branch = self.options.get('source_branch', self.project_config.project__git__default_branch)

        head_branch = repo.branch(source_branch)
        if not repo:
            self.logger.error('Branch {} not found'.format(source_branch))

        # Get existing pull requests targeting a target branch
        existing_prs = []
        for pr in repo.iter_pulls(state='open'):
            if pr.base.ref.startswith(branch_prefix):
                existing_prs.append(pr.base.ref)
        
        targets = []
        for branch in repo.iter_branches():
            if not branch.name.startswith(branch_prefix):
                self.logger.info('Skipping branch {}: does not match prefix {}'.format(branch.name, branch_prefix))
                continue
            if branch.name == source_branch:
                self.logger.info('Skipping branch {}: is source branch'.format(branch.name))
                continue
               
            compare = repo.compare_commits(branch.commit.sha, commit)
            if not compare.files:
                self.logger.info('Skipping branch {}: no file diffs found'.format(branch.name))
                continue
   
            try: 
                result = repo.merge(branch.name, source_branch)
                self.logger.info('Merged {} commits into {}'.format(compare.behind_by, branch.name))
            except GitHubError as e:
                if e.code != 409:
                    raise

                if branch.name in existing_prs:
                    self.logger.info('Merge conflict on branch {}: merge PR already exists'.format(branch.name))
                    continue 

                pull = repo.create_pull(
                    title = 'Merge {} into {}'.format(source_branch, branch.name),
                    base = branch.name,
                    head = source_branch,
                    body = 'This pull request was automatically generated because an automated merge hit a merge conflict',
                )

                self.logger.info('Merge conflict on branch {}: created pull request #{}'.format(branch.name, pull.number))



import fnmatch
import os
import re
import StringIO
import zipfile

import requests

from xml.etree.ElementTree import ElementTree

CUMULUSCI_PATH = os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), '..')
)


def findReplace(find, replace, directory, filePattern, logger=None, max=None):
    for path, dirs, files in os.walk(os.path.abspath(directory)):
        for filename in fnmatch.filter(files, filePattern):
            filepath = os.path.join(path, filename)
            with open(filepath) as f:
                s = f.read()
            if max:
                s_updated = s.replace(find, replace, max)
            else:
                s_updated = s.replace(find, replace)
            if s != s_updated:
                if logger:
                    logger.info('Updating {}'.format(filepath))
                with open(filepath, "w") as f:
                    f.write(s_updated)


def findReplaceRegex(find, replace, directory, filePattern, logger=None):
    for path, dirs, files in os.walk(os.path.abspath(directory)):
        for filename in fnmatch.filter(files, filePattern):
            filepath = os.path.join(path, filename)
            with open(filepath) as f:
                s = f.read()
            s_updated = re.sub(find, replace, s)
            if s != s_updated:
                if logger:
                    logger.info('Updating {}'.format(filepath))
                with open(filepath, "w") as f:
                    f.write(s_updated)


def findRename(find, replace, directory, logger=None):
    for path, dirs, files in os.walk(os.path.abspath(directory)):
        for filename in files:
            filepath = os.path.join(path, filename)
            if logger:
                logger.info('Renaming {}'.format(filepath))
            os.rename(filepath, os.path.join(
                path,
                filename.replace(find, replace),
            ))


def removeXmlElement(name, directory, file_pattern, logger=None):
    for path, dirs, files in os.walk(os.path.abspath(directory)):
        for filename in fnmatch.filter(files, file_pattern):
            filepath = os.path.join(path, filename)
            tree = ElementTree()
            tree.parse(filepath)
            root = tree.getroot()
            remove = root.findall(
                './/{{http://soap.sforce.com/2006/04/metadata}}{}'.format(name)
            )
            if not remove:
                continue

            if logger:
                logger.info(
                    'Modifying {} to remove <{}> elements'.format(
                        filepath,
                        name,
                    )
                )

            parent_map = {c: p for p in tree.iter() for c in p}

            for elem in remove:
                parent = parent_map[elem]
                parent.remove(elem)

            tree.write(
                filepath,
                encoding="UTF-8",
                default_namespace='http://soap.sforce.com/2006/04/metadata',
            )


def download_extract_zip(url, target, subfolder=None):
    resp = requests.get(url)
    zip_content = StringIO.StringIO(resp.content)
    zip_file = zipfile.ZipFile(zip_content)
    if subfolder:
        zip_file = zip_subfolder(zip_file, subfolder)
    zip_file.extractall(target)


def zip_subfolder(zip_src, path):
    if not path.endswith('/'):
        path = path + '/'

    zip_dest = zipfile.ZipFile(StringIO.StringIO(), 'w', zipfile.ZIP_DEFLATED)
    for name in zip_src.namelist():
        if not name.startswith(path):
            continue

        content = zip_src.read(name)
        rel_name = name.replace(path, '', 1)

        if rel_name:
            zip_dest.writestr(rel_name, content)

    return zip_dest


def doc_task(task_name, task_config, project_config=None, org_config=None):
    """ Document a (project specific) task configuration in RST format. """
    from cumulusci.core.utils import import_class
    doc = []
    doc.append(
        '{}\n==========================================\n'.format(task_name))
    doc.append('**Description:** {}\n'.format(task_config.description))
    doc.append('**Class::** {}\n'.format(task_config.class_path))

    task_class = import_class(task_config.class_path)
    if task_class.task_options:
        doc.append('Options:\n------------------------------------------\n')
        defaults = task_config.options
        if not defaults:
            defaults = {}
        for name, option in task_class.task_options.items():
            default = defaults.get(name)
            if default:
                default = ' **Default: {}**'.format(default)
            else:
                default = ''
            if option.get('required'):
                doc.append('* **{}** *(required)*: {}{}'.format(
                    name,
                    option.get('description'),
                    default))
            else:
                doc.append('* **{}**: {}{}'.format(
                    name,
                    option.get('description'),
                    default))

    return '\n'.join(doc)


def package_xml_from_dict(items, api_version, package_name=None):
    lines = []

    # Print header
    lines.append(u'<?xml version="1.0" encoding="UTF-8"?>')
    lines.append(u'<Package xmlns="http://soap.sforce.com/2006/04/metadata">')

    # Include package name if specified
    if package_name:
        lines.append('    <fullName>{}</fullName'.format(package_name))

    # Print types sections
    md_types = items.keys()
    md_types.sort()
    for md_type in md_types:
        members = items[md_type]
        members.sort()
        lines.append('    <types>')
        for member in members:
            lines.append('        <members>{}</members>'.format(member))
        lines.append('        <name>{}</name>'.format(md_type))
        lines.append('    </types>')

    # Print footer
    lines.append(u'    <version>{0}</version>'.format(api_version))
    lines.append(u'</Package>')

    return u'\n'.join(lines)



# coding=utf-8

import datetime
import httplib
import json
import os
import unittest

import responses

from cumulusci.tasks.release_notes.generator import BaseReleaseNotesGenerator
from cumulusci.tasks.release_notes.generator import StaticReleaseNotesGenerator
from cumulusci.tasks.release_notes.generator import DirectoryReleaseNotesGenerator
from cumulusci.tasks.release_notes.generator import GithubReleaseNotesGenerator
from cumulusci.tasks.release_notes.generator import PublishingGithubReleaseNotesGenerator
from cumulusci.tasks.release_notes.parser import BaseChangeNotesParser
from cumulusci.tasks.release_notes.tests.util_github_api import GithubApiTestMixin

__location__ = os.path.split(os.path.realpath(__file__))[0]


class DummyParser(BaseChangeNotesParser):

    def parse(self, change_note):
        pass

    def _render(self):
        return 'dummy parser output'.format(self.title)


class TestBaseReleaseNotesGenerator(unittest.TestCase):

    def test_render_no_parsers(self):
        release_notes = BaseReleaseNotesGenerator()
        content = release_notes.render()
        self.assertEqual(content, '')

    def test_render_dummy_parsers(self):
        release_notes = BaseReleaseNotesGenerator()
        release_notes.parsers.append(DummyParser('Dummy 1'))
        release_notes.parsers.append(DummyParser('Dummy 2'))
        expected = u'# Dummy 1\r\n\r\ndummy parser output\r\n\r\n' +\
                   u'# Dummy 2\r\n\r\ndummy parser output'
        self.assertEqual(release_notes.render(), expected)


class TestStaticReleaseNotesGenerator(unittest.TestCase):

    def test_init_parser(self):
        release_notes = StaticReleaseNotesGenerator([])
        assert len(release_notes.parsers) == 3


class TestDirectoryReleaseNotesGenerator(unittest.TestCase):

    def test_init_parser(self):
        release_notes = DirectoryReleaseNotesGenerator('change_notes')
        assert len(release_notes.parsers) == 3

    def test_full_content(self):
        change_notes_dir = os.path.join(
            __location__,
            'change_notes',
            'full',
        )
        release_notes = DirectoryReleaseNotesGenerator(
            change_notes_dir,
        )

        content = release_notes()
        expected = "# Critical Changes\r\n\r\n* This will break everything!\r\n\r\n# Changes\r\n\r\nHere's something I did. It was really cool\r\nOh yeah I did something else too!\r\n\r\n# Issues Closed\r\n\r\n#2345\r\n#6236"
        print expected
        print '-------------------------------------'
        print content

        self.assertEqual(content, expected)


class TestGithubReleaseNotesGenerator(unittest.TestCase):

    def setUp(self):
        self.current_tag = 'prod/1.4'
        self.last_tag = 'prod/1.3'
        self.github_info = {
            'github_owner': 'TestOwner',
            'github_repo': 'TestRepo',
            'github_username': 'TestUser',
            'github_password': 'TestPass',
        }

    def test_init_without_last_tag(self):
        github_info = self.github_info.copy()
        generator = GithubReleaseNotesGenerator(github_info, self.current_tag)
        self.assertEqual(generator.github_info, github_info)
        self.assertEqual(generator.current_tag, self.current_tag)
        self.assertEqual(generator.last_tag, None)
        self.assertEqual(generator.change_notes.current_tag, self.current_tag)
        self.assertEqual(generator.change_notes._last_tag, None)

    def test_init_with_last_tag(self):
        github_info = self.github_info.copy()
        generator = GithubReleaseNotesGenerator(
            github_info, self.current_tag, self.last_tag)
        self.assertEqual(generator.github_info, github_info)
        self.assertEqual(generator.current_tag, self.current_tag)
        self.assertEqual(generator.last_tag, self.last_tag)
        self.assertEqual(generator.change_notes.current_tag, self.current_tag)
        self.assertEqual(generator.change_notes._last_tag, self.last_tag)


class TestPublishingGithubReleaseNotesGenerator(unittest.TestCase, GithubApiTestMixin):

    def setUp(self):
        self.init_github()
        self.github_info = {
            'github_owner': 'TestOwner',
            'github_repo': 'TestRepo',
            'github_username': 'TestUser',
            'github_password': 'TestPass',
        }

    @responses.activate
    def test_publish_update_unicode(self):
        tag = 'prod/1.4'
        note = u'“Unicode quotes”'
        self._mock_release(False, tag, True, '')
        # create generator
        generator = self._create_generator(tag)
        # inject content into Changes parser
        generator.parsers[1].content.append(note)
        # render and publish
        content = generator.render()
        release_body = generator.publish(content)
        # verify
        expected_release_body = u'# Changes\r\n\r\n{}'.format(note)
        self.assertEqual(release_body, expected_release_body)
        response_body = json.loads(responses.calls._calls[1].request.body)
        self.assertEqual(response_body['draft'], True)
        self.assertEqual(response_body['prerelease'], False)
        self.assertEqual(len(responses.calls._calls), 2)

    @responses.activate
    def test_publish_update_no_body(self):
        tag = 'prod/1.4'
        self._mock_release(False, tag, True, '')
        # create generator
        generator = self._create_generator(tag)
        # inject content into Changes parser
        generator.parsers[1].content.append('foo')
        # render and publish
        content = generator.render()
        release_body = generator.publish(content)
        # verify
        expected_release_body = '# Changes\r\n\r\nfoo'
        self.assertEqual(release_body, expected_release_body)

    @responses.activate
    def test_publish_update_content_before(self):
        tag = 'prod/1.4'
        self._mock_release(False, tag, True, 'foo\n# Changes\nbar')
        # create generator
        generator = self._create_generator(tag)
        # inject content into parser
        generator.parsers[1].content.append('baz')
        # render and publish
        content = generator.render()
        release_body = generator.publish(content)
        # verify
        self.assertEqual(release_body, 'foo\r\n# Changes\r\n\r\nbaz')

    @responses.activate
    def test_publish_update_content_after(self):
        tag = 'prod/1.4'
        self._mock_release(False, tag, True, '# Changes\nbar\n# Foo\nfoo')
        # create generator
        generator = self._create_generator(tag)
        # inject content into parser
        generator.parsers[1].content.append('baz')
        # render and publish
        content = generator.render()
        release_body = generator.publish(content)
        # verify
        self.assertEqual(release_body, '# Changes\r\n\r\nbaz\r\n\r\n# Foo\r\nfoo')

    @responses.activate
    def test_publish_update_content_before_and_after(self):
        tag = 'prod/1.4'
        self._mock_release(False, tag, True, 'foo\n# Changes\nbar\n# Foo\nfoo')
        # create generator
        generator = self._create_generator(tag)
        # inject content into parser
        generator.parsers[1].content.append('baz')
        # render and publish
        content = generator.render()
        release_body = generator.publish(content)
        # verify
        self.assertEqual(release_body,
            'foo\r\n# Changes\r\n\r\nbaz\r\n\r\n# Foo\r\nfoo')

    @responses.activate
    def test_publish_update_content_between(self):
        tag = 'prod/1.4'
        self._mock_release(False, tag, True,
            '# Critical Changes\nbar\n# Foo\nfoo\n# Changes\nbiz')
        # create generator
        generator = self._create_generator(tag)
        # inject content into parser
        generator.parsers[0].content.append('faz')
        generator.parsers[1].content.append('fiz')
        # render and publish
        content = generator.render()
        release_body = generator.publish(content)
        # verify
        self.assertEqual(release_body,
            '# Critical Changes\r\n\r\nfaz\r\n\r\n' +
            '# Foo\r\nfoo\r\n# Changes\r\n\r\nfiz')

    @responses.activate
    def test_publish_update_content_before_after_and_between(self):
        tag = 'prod/1.4'
        self._mock_release(False, tag, True,
            'goo\n# Critical Changes\nbar\n# Foo\nfoo\n' +
            '# Changes\nbiz\n# Zoo\nzoo')
        # create generator
        generator = self._create_generator(tag)
        # inject content into parser
        generator.parsers[0].content.append('faz')
        generator.parsers[1].content.append('fiz')
        # render and publish
        content = generator.render()
        release_body = generator.publish(content)
        # verify
        self.assertEqual(release_body,
            'goo\r\n# Critical Changes\r\n\r\nfaz\r\n\r\n' +
            '# Foo\r\nfoo\r\n# Changes\r\n\r\nfiz\r\n\r\n' +
            '# Zoo\r\nzoo')

    def _create_generator(self, current_tag, last_tag=None):
        generator = PublishingGithubReleaseNotesGenerator(
            self.github_info.copy(), current_tag, last_tag)
        return generator

    def _mock_release(self, beta, tag, update, body):
        if beta:
            draft = False
            prerelease = True
        else:
            draft = True
            prerelease = False
        # mock the attempted GET of non-existent release
        api_url = '{}/releases/tags/{}'.format(self.repo_api_url, tag)
        if update:
            expected_response = self._get_expected_release(
                body, draft, prerelease)
            status = httplib.OK
        else:
            expected_response = self._get_expected_not_found()
            status = httplib.NOT_FOUND
        responses.add(
            method=responses.GET,
            url=api_url,
            json=expected_response,
            status=status,
        )
        # mock the release creation
        api_url = '{}/releases'.format(self.repo_api_url)
        expected_response = self._get_expected_release(None, draft, prerelease)
        responses.add(
            method=responses.POST,
            url=api_url,
            json=expected_response,
        )



import time
from datetime import datetime

from cumulusci.core.tasks import BaseTask
from cumulusci.tasks.salesforce import BaseSalesforceApiTask
from cumulusci.tasks.push.push_api import SalesforcePushApi

class BaseSalesforcePushTask(BaseSalesforceApiTask):
    completed_statuses = ['Succeeded','Failed','Cancelled']
    api_version = '38.0'

    def _init_task(self):
        super(BaseSalesforcePushTask, self)._init_task()
        self.push = SalesforcePushApi(self.sf, self.logger)

    def _parse_version(self, version):
        # Parse the version number string
        major = None
        minor = None
        patch = None
        build = None
        state = 'Released'
        version_parts = version.split('.')
        if len(version_parts) >= 1:
            major = version_parts[0]
        if len(version_parts) == 2:
            minor = version_parts[1]
            if minor.find('Beta') != -1:
                state = 'Beta'
                minor, build = minor.replace(' (Beta ',',').replace(')','').split(',')
        if len(version_parts) > 2:
            minor = version_parts[1]
            patch = version_parts[2]
            if patch.find('Beta') != -1:
                state = 'Beta'
                patch, build = minor.replace(' (Beta ',',').replace(')','').split(',')
        
        return {
            'major': major,
            'minor': minor,
            'patch': patch,
            'build': build,
            'state': state,
        }

    def _get_version(self, package, version):

        version_info = self._parse_version(version)

        version_where = "ReleaseState = '{}' AND MajorVersion = {} AND MinorVersion = {}".format(
            version_info['state'],
            version_info['major'],
            version_info['minor'],
        )
        if version_info.get('patch'):
             version_where += " AND PatchVersion = {}".format(version_info['patch'])
        if version_info['state'] == 'Beta' and version_info.get('build'):
             version_where += " AND BuildNumber = {}".format(version_info['build'])

        version = package.get_package_version_objs(version_where, limit=1)
        if not version:
            raise PushApiObjectNotFound('PackageVersion not found.  Namespace = {}, Version Info = {}'.format(package.namespace, version_info))
        return version[0]


    def _get_package(self, namespace):
        package = self.push.get_package_objs(
            "NamespacePrefix = '{}'".format(namespace),
            limit=1
        )
        
        if not package:
            raise PushApiObjectNotFound('The package with namespace {} was not found'.format(namespace))

        return package[0]

    def _load_orgs_file(self, path):
        orgs = []
        with open(path, 'r') as f_orgs:
            for org in f_orgs:
                orgs.append(org.strip())
        return orgs

    def _report_push_status(self, request_id):
        default_where = {'PackagePushRequest': "Id = '{}'".format(request_id)}

        # Create a new PushAPI instance with different settings than self.push
        self.push_report = SalesforcePushApi(
            self.sf, 
            self.logger, 
            lazy = ['subscribers','jobs'],
            default_where = default_where,
        )

        # Get the push request
        push_request = self.push_report.get_push_request_objs("Id = '{}'".format(request_id), limit=1)
        if not push_request:
            raise PushApiObjectNotFound('Push Request {} was not found'.format(push_request))
        push_request = push_request[0]


        # Check if the request is complete
        interval = 10
        if push_request.status not in self.completed_statuses:
            self.logger.info(
                'Push request is not yet complete.  Polling for status every {} seconds until completion...'.format(interval)
            )

        # Loop waiting for request completion
        i = 0 
        while push_request.status not in self.completed_statuses:
            if i == 10:
                self.logger.info('This is taking a while! Polling every 60 seconds...')
                interval = 60
            time.sleep(interval)

            # Clear the method level cache on get_push_requests and get_push_request_objs
            self.push_report.get_push_requests.cache.clear()
            self.push_report.get_push_request_objs.cache.clear()

            # Get the push_request again
            push_request = self.push_report.get_push_request_objs("Id = '{}'".format(request_id), limit=1)[0]

            self.logger.info(push_request.status)

            i += 1

        failed_jobs = []
        success_jobs = []
        cancelled_jobs = []

        jobs = push_request.get_push_job_objs()
        for job in jobs:
            if job.status == 'Failed':
                failed_jobs.append(job)
            elif job.status == 'Succeeded':
                success_jobs.append(job)
            elif job.status == 'Cancelled':
                cancelled_jobs.append(job)

        self.logger.info(
            "Push complete: {} succeeded, {} failed, {} cancelled".format(
                len(success_jobs),
                len(failed_jobs),
                len(cancelled_jobs),
            )
        )

        failed_by_error = {}
        for job in failed_jobs:
            errors = job.get_push_error_objs()
            for error in errors:
                error_key = (error.error_type, error.title, error.message, error.details)
                if error_key not in failed_by_error:
                    failed_by_error[error_key] = []
                failed_by_error[error_key].append(error)

        if failed_jobs:
            self.logger.info("-----------------------------------")
            self.logger.info("Failures by error type")
            self.logger.info("-----------------------------------")
            for key, errors in failed_by_error.items():
                self.logger.info("    ")
                self.logger.info("{} failed with...".format(len(errors)))
                self.logger.info("    Error Type = {}".format(key[0]))
                self.logger.info("    Title = {}".format(key[1]))
                self.logger.info("    Message = {}".format(key[2]))
                self.logger.info("    Details = {}".format(key[3]))
        
    
class SchedulePushOrgList(BaseSalesforcePushTask):

    task_options = {
        'orgs': {
            'description': "The path to a file containing one OrgID per line.",
            'required': True,
        },
        'version': {
            'description': "The managed package version to push",
            'required': True,
        },
        'namespace': {
            'description': "The managed package namespace to push. Defaults to project__package__namespace.",
        },
        'start_time': {
            'description': "Set the start time (UTC) to queue a future push. Ex: 2016-10-19T10:00",
        },
        'batch_size': {
            'description': "Set batch size used to add individual orgs to the push request.  Defaults to 200",
        },
    }

    def _init_task(self):
        super(SchedulePushOrgList, self)._init_task()
        self.push = SalesforcePushApi(self.sf, self.logger, self.options['batch_size'])

    def _init_options(self, kwargs):
        super(SchedulePushOrgList, self)._init_options(kwargs)

        # Set the namespace option to the value from cumulusci.yml if not already set
        if not 'namespace' in self.options:
            self.options['namespace'] = self.project_config.project__package__namespace
        if not 'batch_size' in self.options:
            self.options['batch_size'] = 200

    def _get_orgs(self):
        return self._load_orgs_file(self.options.get('orgs'))

    def _run_task(self):
        orgs = self._get_orgs()
        package = self._get_package(self.options.get('namespace')) 
        version = self._get_version(package, self.options.get('version'))

        start_time = self.options.get('start_time')
        if start_time:
            start_time = datetime.strptime(start_time, "%Y-%m-%dT%H:%M") # Example: 2016-10-19T10:00

        self.request_id = self.push.create_push_request(version, orgs, start_time)

        if len(orgs) > 1000:
            self.logger.info("Delaying 30 seconds to allow all jobs to initialize...")
            time.sleep(30)

        self.logger.info('Setting status to Pending to queue execution.')
        if start_time:
            self.logger.info('The push request will start at {}'.format(start_time))

        # Run the job
        self.logger.info(self.push.run_push_request(self.request_id))
        self.logger.info('Push Request {} is queued for execution.'.format(self.request_id))

        # Report the status
        self._report_push_status(self.request_id)

class SchedulePushOrgQuery(SchedulePushOrgList):
    task_options = {
        'version': {
            'description': "The managed package version to push",
            'required': True,
        },
        'subscriber_where': {
            'description': "A SOQL style where clause for filtering PackageSubscriber objects.  Ex: OrgType = 'Sandbox'",
        },
        'min_version': {
            'description': "If set, no subscriber with a version lower than min_version will be selected for push",
        },
        'namespace': {
            'description': "The managed package namespace to push. Defaults to project__package__namespace.",
        },
        'start_time': {
            'description': "Set the start time (UTC) to queue a future push. Ex: 2016-10-19T10:00",
        }
    }

    def _get_orgs(self):
        subscriber_where = self.options.get('subscriber_where')
        default_where = {'PackageSubscriber': "OrgStatus = 'Active' AND InstalledStatus = 'i'"}
        if subscriber_where:
            default_where['PackageSubscriber'] += " AND ({})".format(subscriber_where)

        push_api = SalesforcePushApi(self.sf, self.logger, default_where=default_where.copy())

        package = self._get_package(self.options.get('namespace')) 
        version = self._get_version(package, self.options.get('version'))
        min_version = self.options.get('min_version')
        if min_version:
            min_version = self._get_version(package, self.options.get('min_version'))

        orgs = []

        if min_version:
            # If working with a range of versions, use an inclusive search
            versions = version.get_older_released_version_objs(greater_than_version=min_version)
            included_versions = []
            for include_version in versions:
                included_versions.append(str(include_version.sf_id))
            if not included_versions:
                raise ValueError(
                    'No versions found between version id {} and {}'.format(
                        version.version_number, min_version.version_number
                    )
                )

            # Query orgs for each version in the range individually to avoid query timeout errors with querying multiple versions
            for included_version in included_versions:
                # Clear the get_subscribers method cache before each call
                push_api.get_subscribers.cache.clear()
                push_api.default_where['PackageSubscriber'] = "{} AND MetadataPackageVersionId = '{}'".format(
                    default_where['PackageSubscriber'],
                    included_version,
                )
                for subscriber in push_api.get_subscribers():
                    orgs.append(subscriber['OrgKey'])

        else:
            # If working with a specific version rather than a range, use an exclusive search
            # Add exclusion of all orgs running on newer releases
            newer_versions = version.get_newer_released_version_objs()
            excluded_versions = [str(version.sf_id),]
            for newer in newer_versions:
                excluded_versions.append(str(newer.sf_id))
            if len(excluded_versions) == 1:
                push_api.default_where['PackageSubscriber'] += " AND MetadataPackageVersionId != '{}'".format(
                    excluded_versions[0],
                )
            else:
                push_api.default_where['PackageSubscriber'] += " AND MetadataPackageVersionId NOT IN {}".format(
                    "('" + "','".join(excluded_versions) + "')"
                )

            for subscriber in push_api.get_subscribers():
                orgs.append(subscriber['OrgKey'])

        return orgs


class GetSubscriberList(BaseSalesforceApiTask):
    """ Get subscribed org info and write to local CSV file """
    task_options = {
        'filename': {
            'description': 'File where org IDs will be written',
            'required': True,
        },
    }

    def _run_task(self):
        raise NotImplementedError


class FilterSubscriberList(BaseTask):
    """ Filter subscriber org list by org type, version. Write file with org IDs only """
    task_options = {
        'file_in': {
            'description': 'CSV file with full org info',
            'required': True,
        },
        'file_out': {
            'description': 'File where org IDs will be written',
            'required': True,
        },
        'org_type': {
            'description': 'Filter by org type'
        },
        'version': {
            'description': 'Filter by installed package version'
        },
    }

    def _run_task(self):
        raise NotImplementedError
